import sys
from argparse import ArgumentParser

from lark import UnexpectedCharacters
from lark.exceptions import LarkError

from .cube_builder import init_cube_args_parser, build_cube
from .error_display import ErrorsOutput
from .options import file_contents_type
from .postprocessors_builder import init_postprocessors_args_parser, build_postprocessors_chain
from ..compiler import Stack, parser
from ..compiler.errors import CompileTimeError, FunctionArgumentsError
from ..cube_runtime import CubeRuntime
from ..execution import ExecutionContext
from ..stdlib import stdlib
from .. import __version__


def main():
    args_parser = ArgumentParser()
    args_parser.add_argument("source", type=file_contents_type,
                             help="program's source")
    args_parser.add_argument("-v", "--version", action="version", version="%(prog)s " + __version__)

    init_cube_args_parser(args_parser)
    init_postprocessors_args_parser(args_parser)
    args = args_parser.parse_args()

    cube, orientation = build_cube(args)
    postprocessor = build_postprocessors_chain(args)
    runtime = CubeRuntime(cube, orientation, postprocessor.process, postprocessor.done)

    stack = Stack()
    stdlib.initialize_stack(stack)
    runtime.functions.initialize_stack(stack)
    exec_globals = {**stdlib.exec_globals, **runtime.functions.exec_globals}
    context = ExecutionContext(exec_globals)

    errors = ErrorsOutput(sys.stderr, use_color=True)
    try:
        program = parser.parse(args.source, stack)
        context.compile(program)
    except UnexpectedCharacters as e:
        errors.write_error(f"Unexpected character: `{args.source[e.pos_in_stream]}`", e.line - 1, e.column - 1)
        errors.display_code(args.source, e.line - 1, e.column - 1, e.line - 1, e.column - 1)
        return
    except LarkError as e:
        errors.write_error(str(e))
        return
    except CompileTimeError as e:
        message = str(e)
        errors.write_error(message, e.start_line, e.start_column)
        if isinstance(e, FunctionArgumentsError):
            errors.write_supplied_arguments(e.arguments)
            errors.write_function_overloads(e.function_name, e.function)

        end_line = e.end_line if e.end_line is not None else e.start_line
        end_column = e.end_column if e.end_column is not None else e.start_column
        errors.display_code(args.source, e.start_line - 1, e.start_column - 1, end_line - 1, end_column - 1)
        return

    # if not pycode:
    context.execute(errors)
    runtime.finished()



from typing import NamedTuple, List, Tuple, Union, Callable, Optional

from .expression import TemplateType
from .types import Type, Bool, Real, Integer, Void

ArgumentsType = Union[List[Tuple[Tuple[Type, Type], Type]],
                      Callable[[Type, Type], Optional[Type]]]


class BinaryOperator(NamedTuple):
    symbol: str
    expression: TemplateType
    arguments: ArgumentsType


def _operator_equals(a: Type, b: Type) -> Optional[Type]:
    if a == Void or a != b:
        return None
    else:
        return Bool


BINARY_OPERATORS = [[
    BinaryOperator("xor", ["(", 0, ") != (", 1, ")"], [((Bool, Bool), Bool)]),
], [
    BinaryOperator("or", ["(", 0, ") or (", 1, ")"], [((Bool, Bool), Bool)]),
], [
    BinaryOperator("and", ["(", 0, ") and (", 1, ")"], [((Bool, Bool), Bool)]),
], [
    BinaryOperator("==", ["(", 0, ") == (", 1, ")"], _operator_equals),
    BinaryOperator("!=", ["(", 0, ") != (", 1, ")"], _operator_equals)
], [
    BinaryOperator("<", ["(", 0, ") < (", 1, ")"], [((Real, Real), Bool)]),
    BinaryOperator(">", ["(", 0, ") > (", 1, ")"], [((Real, Real), Bool)]),
    BinaryOperator("<=", ["(", 0, ") <= (", 1, ")"], [((Real, Real), Bool)]),
    BinaryOperator(">=", ["(", 0, ") >= (", 1, ")"], [((Real, Real), Bool)])
], [
    BinaryOperator("+", ["(", 0, ") + (", 1, ")"], [((Integer, Integer), Integer), ((Real, Real), Real)]),
    BinaryOperator("-", ["(", 0, ") - (", 1, ")"], [((Integer, Integer), Integer), ((Real, Real), Real)])
], [
    BinaryOperator("*", ["(", 0, ") * (", 1, ")"], [((Integer, Integer), Integer), ((Real, Real), Real)]),
    BinaryOperator("/", ["(", 0, ") / (", 1, ")"], [((Real, Real), Real)]),
    BinaryOperator("%", ["(", 0, ") % (", 1, ")"], [((Integer, Integer), Integer)])
]]


def operator_applicable(actual_types: List[Type],
                        operator_types: ArgumentsType) -> Optional[Type]:
    if isinstance(operator_types, list):
        for operand_types, result_type in operator_types:
            if all(ar1.is_assignable(ar2) for ar1, ar2 in zip(operand_types, actual_types)):
                return result_type
        else:
            return None
    else:
        return operator_types(*actual_types)



from typing import Deque
from collections import deque

from ..actions import Action, Rotate, Turn
from .base import PostprocessorBase


class OptimizingPostprocessor(PostprocessorBase[Action, Action]):
    def __init__(self):
        super().__init__()
        self.stack: Deque[Action] = deque()

    def process(self, action: Action):
        if len(self.stack) == 0 or type(self.stack[-1]) != type(action):
            self.stack.append(action)
            return
        on_top = self.stack[-1]
        if isinstance(action, Rotate):
            on_top: Rotate = on_top
            turns = 2 if on_top.twice else 1
            if on_top.axis_side == action.axis_side:
                turns += 2 if action.twice else 1
            elif on_top.axis_side == action.axis_side.opposite():
                turns -= 2 if action.twice else 1
            else:
                self.stack.append(action)
                return

            self.stack.pop()
            turns = (4 + turns) % 4
            if turns != 0:
                if turns == 3:
                    self.stack.append(Rotate(on_top.axis_side.opposite(), False))
                else:
                    self.stack.append(Rotate(on_top.axis_side, turns == 2))
        elif isinstance(action, Turn):
            on_top: Turn = on_top
            if on_top.type != action.type or on_top.indices != action.indices:
                self.stack.append(action)
                return
            self.stack.pop()
            turns = (action.turns + on_top.turns) % 4
            if turns != 0:
                self.stack.append(Turn(action.type, action.indices, turns))
        else:
            self.stack.append(action)

    def done(self):
        for item in self.stack:
            self._return(item)
        super(OptimizingPostprocessor, self).done()



from abc import ABC, abstractmethod
from typing import Tuple, List, Generic, TypeVar, Optional

from .orientation import Color

T = TypeVar("T")


class Component(Generic[T]):
    def __init__(self, color: Color, data: Optional[T]) -> None:
        self.color: Color = color
        self.data: Optional[T] = data

    def __repr__(self) -> str:
        return f"Component({self.color}, {self.data})"


class ICubeSide(ABC, Generic[T]):
    def __init__(self) -> None:
        self.colors = ColorsAccessor(self)

    @property
    @abstractmethod
    def rows(self) -> int:
        pass

    @property
    @abstractmethod
    def columns(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: Tuple[int, int]) -> Component[T]:
        pass

    @abstractmethod
    def __setitem__(self, key: Tuple[int, int], value: Component[T]) -> None:
        pass

    def create_view(self, rotation: int) -> "CubeSideView":
        return CubeSideView(self, rotation)

    def rotate(self, amount: int) -> None:
        amount = amount % 4
        if amount == 0:
            return
        elif amount % 2 == 0:
            for i in range(self.rows // 2):
                for j in range(self.columns):
                    i1, i2 = self.rows - 1 - i, self.columns - 1 - j
                    self[i, j], self[i1, i2] = self[i1, i2], self[i, j]
            if self.rows % 2 != 0:
                for j in range(self.columns // 2):
                    i = self.rows // 2
                    self[i, j], self[i, self.columns - i] = self[i, self.columns - i], self[i, j]
        else:
            if self.rows != self.columns:
                raise AttributeError("Cannot rotate a side: it would have different shape after rotation")
            size = self.rows
            for i in range(size // 2):
                for j in range(i, size - 1 - i):
                    indices = [(i, j), (j, size - 1 - i), (size - 1 - i, size - 1 - j), (size - 1 - j, i)]
                    values = [self[i1, j1] for i1, j1 in indices]

                    if amount == 3:
                        values = values[1:] + [values[0]]
                    else:
                        values = [values[-1]] + values[:-1]
                    for (i1, j1), val in zip(indices, values):
                        self[i1, j1] = val

    def get_row(self, i: int) -> List[Component[T]]:
        return [self[i, j] for j in range(self.columns)]

    def get_column(self, j: int) -> List[Component[T]]:
        return [self[i, j] for i in range(self.rows)]

    def set_row(self, i: int, values: List[Component[T]]) -> None:
        assert len(values) == self.columns
        for j in range(self.columns):
            self[i, j] = values[j]

    def set_column(self, j: int, values: List[Component[T]]) -> None:
        assert len(values) == self.rows
        for i in range(self.rows):
            self[i, j] = values[i]

    def __repr__(self) -> str:
        content = "/".join(
            ", ".join(f"{self[r, c].color.name[0]}({self[r, c].data})" for c in range(self.columns))
            for r in range(self.rows))
        return f"<Side: {content}>"


class ColorsAccessor(Generic[T]):
    def __init__(self, side: ICubeSide[T]) -> None:
        self.side: ICubeSide[T] = side

    def __repr__(self):
        return "/".join("".join(self[i, j].name[0] for j in range(self.side.columns))
                        for i in range(self.side.rows))

    def __getitem__(self, item: Tuple[int, int]) -> Color:
        return self.side[item].color

    def __setitem__(self, key: Tuple[int, int], value: Color) -> None:
        self.side[key].color = value


class CubeSideView(ICubeSide[T]):
    def __init__(self, side: ICubeSide, rotation: int) -> None:
        super().__init__()
        self.side: ICubeSide = side
        self.rotation: int = rotation % 4

    @property
    def rows(self) -> int:
        return self.side.rows if self.rotation % 2 == 0 else self.side.columns

    @property
    def columns(self) -> int:
        return self.side.columns if self.rotation % 2 == 0 else self.side.rows

    def _transform_coord(self, coords: Tuple[int, int]) -> Tuple[int, int]:
        i, j = coords
        if self.rotation == 1:
            return j, self.side.columns - 1 - i
        elif self.rotation == 2:
            return self.side.rows - 1 - i, self.side.columns - 1 - j
        elif self.rotation == 3:
            return self.side.rows - 1 - j, i
        else:
            return i, j

    def __getitem__(self, item: Tuple[int, int]) -> Component[T]:
        i, j = self._transform_coord(item)
        return self.side[i, j]

    def __setitem__(self, key: Tuple[int, int], value: Component[T]) -> None:
        i, j = self._transform_coord(key)
        self.side[i, j] = value


class CubeSide(ICubeSide[T]):
    def __init__(self, rows: int, columns: int, default: Color):
        super().__init__()
        self.shape = (rows, columns)
        self.items: List[List[Component[T]]] = [[Component[T](default, None) for _j in range(columns)]
                                                for _i in range(rows)]

    def __getitem__(self, item: Tuple[int, int]) -> Component[T]:
        i, j = item
        return self.items[i][j]

    def __setitem__(self, key: Tuple[int, int], value: Component[T]) -> None:
        i, j = key
        self.items[i][j] = value

    @property
    def rows(self) -> int:
        return self.shape[0]

    @property
    def columns(self) -> int:
        return self.shape[1]



from typing import Callable
from .rt_error import RuntimeError


class RuntimeFunction:
    def __init__(self, name: str, function: Callable):
        self._function = function
        self.name = name

    def __call__(self, *args, **kwargs):
        try:
            return self._function(*args, **kwargs)
        except Exception as e:
            raise RuntimeError.update_error(self.name, e)


def runtime_function(name: str):
    return lambda func: RuntimeFunction(name, func)



from typing import List, Union, Dict, Optional, Iterable, Tuple
from .orientation import Color
from .cube import ICubeSide


class Pattern:
    _COLOR_CHARS = {"R": Color.RED, "G": Color.GREEN, "B": Color.BLUE,
                    "W": Color.WHITE, "O": Color.ORANGE, "Y": Color.YELLOW}

    def __init__(self, values: List[List[Union[Color, str, None]]]):
        self.values = values
        self.rows = len(values)
        self.columns = len(values[0])

    def _get_items(self) -> Iterable[Tuple[int, int, Union[str, Color]]]:
        yield from ((row_index, col_index, column)
                    for row_index, row in enumerate(self.values)
                    for col_index, column in enumerate(row)
                    if column is not None)

    def match(self, side: ICubeSide, known: Dict[str, Color]) -> \
            Optional[Dict[str, Color]]:
        if self.rows != side.rows or self.columns != side.columns:
            return None

        result = dict()
        for row, column, cell in self._get_items():
            color = side.colors[row, column]
            if isinstance(cell, Color):
                if cell != color:
                    return None
            elif (cell in known and known[cell] != color) or (cell in result and result[cell] != color):
                return None
            elif cell not in known:
                if color in known.values() or (color in result.values() and result.get(cell, None) != color):
                    return None
                result[cell] = color

        return result



from .optimizer import OptimizingPostprocessor
from .formatter import FormattingPostprocessor
from .orientation_freezer import OrientationFreezePostprocessor
from .base import chain, PostprocessorBase



import enum
from typing import Any, Iterable, Optional, Callable, Tuple, TypeVar
from itertools import groupby
from functools import wraps

T = TypeVar("T")


def count_occurrences(seq: Iterable[T]) -> Iterable[Tuple[T, int]]:
    return ((val, sum(1 for _ in group)) for val, group in groupby(seq))


def pipe(g):
    def wrapper(f):
        @wraps(f)
        def function(*args, **kwargs):
            return g(f(*args, **kwargs))
        return function
    return wrapper


class Color (enum.Enum):
    RED = enum.auto()
    BLUE = enum.auto()
    YELLOW = enum.auto()
    WHITE = enum.auto()
    GREEN = enum.auto()
    ORANGE = enum.auto()


class Side (enum.Enum):
    FRONT = enum.auto()
    BACK = enum.auto()
    LEFT = enum.auto()
    RIGHT = enum.auto()
    TOP = enum.auto()
    BOTTOM = enum.auto()

    def opposite(self):
        opposites = {Side.LEFT: Side.RIGHT, Side.RIGHT: Side.LEFT,
                     Side.TOP: Side.BOTTOM, Side.BOTTOM: Side.TOP,
                     Side.FRONT: Side.BACK, Side.BACK: Side.FRONT}
        return opposites[self]


def _normalize_turns(turns: Iterable[Side]) -> Iterable[Side]:
    for t, count in count_occurrences(turns):
        count = count % 4
        if count == 0:
            continue

        if t not in {Side.RIGHT, Side.TOP, Side.FRONT}:
            count = 4 - count
            t = t.opposite()

        yield from (t for _ in range(count))


class Orientation:
    _RELATIVE_SIDES = {
        Side.FRONT: [Side.TOP, Side.RIGHT, Side.BOTTOM, Side.LEFT],
        Side.LEFT: [Side.TOP, Side.FRONT, Side.BOTTOM, Side.BACK],
        Side.BACK: [Side.TOP, Side.LEFT, Side.BOTTOM, Side.RIGHT],
        Side.RIGHT: [Side.TOP, Side.BACK, Side.BOTTOM, Side.FRONT],
        Side.TOP: [Side.BACK, Side.RIGHT, Side.FRONT, Side.LEFT],
        Side.BOTTOM: [Side.LEFT, Side.FRONT, Side.RIGHT, Side.BACK]
    }

    def __init__(self, front: Side = Side.FRONT, top: Side = Side.TOP) -> None:
        self.front: Side = front
        self.top: Side = top

    @property
    def to_top(self) -> "Orientation":
        return Orientation(self.top, self.front.opposite())

    @property
    def to_left(self) -> "Orientation":
        index = Orientation._RELATIVE_SIDES[self.front].index(self.top)
        index = (4 + index - 1) % 4
        return Orientation(Orientation._RELATIVE_SIDES[self.front][index], self.top)

    @property
    def to_right(self) -> "Orientation":
        index = Orientation._RELATIVE_SIDES[self.front].index(self.top)
        index = (4 + index + 1) % 4
        return Orientation(Orientation._RELATIVE_SIDES[self.front][index], self.top)

    @property
    def to_bottom(self) -> "Orientation":
        return Orientation(self.top.opposite(), self.front)

    def get_side_rotation(self) -> int:
        return Orientation._RELATIVE_SIDES[self.front].index(self.top)

    def __eq__(self, other: Any) -> bool:
        return (isinstance(other, Orientation) and
                other.front == self.front and other.top == self.top)

    def __hash__(self) -> int:
        return hash((self.front, self.top))

    def rotate_clockwise(self) -> "Orientation":
        index = (4 + Orientation._RELATIVE_SIDES[self.front].index(self.top) - 1) % 4
        return Orientation(self.front, Orientation._RELATIVE_SIDES[self.front][index])

    def rotate_counterclockwise(self) -> "Orientation":
        index = (Orientation._RELATIVE_SIDES[self.front].index(self.top) + 1) % 4
        return Orientation(self.front, Orientation._RELATIVE_SIDES[self.front][index])

    def __repr__(self) -> str:
        return f"Orientation(front={repr(self.front)}, top={repr(self.top)})"

    @classmethod
    def regular(cls, side: Side) -> "Orientation":
        if side == Side.BOTTOM:
            return Orientation(side, Side.FRONT)
        else:
            return Orientation(side, Orientation._RELATIVE_SIDES[side][0])

    def iterate_rotations(self, keeping: Optional[Side] = None) -> Iterable["Orientation"]:
        def orientation_changes(orientation: Orientation, action: Callable[[Orientation], Orientation]):
            for _ in range(4):
                yield orientation
                orientation = action(orientation)

        def iterate_cube_rotations() -> Iterable["Orientation"]:
            if keeping == Side.FRONT or keeping == Side.BACK:
                yield from orientation_changes(self, lambda x: x.rotate_clockwise())
            elif keeping == Side.LEFT or keeping == Side.RIGHT:
                yield from orientation_changes(self, lambda x: x.to_top)
            else:
                yield from orientation_changes(self, lambda x: x.to_left)
            if keeping is None:
                yield self.to_top
                yield self.to_bottom

        if keeping is not None:
            yield from iterate_cube_rotations()
        else:
            for rotation in iterate_cube_rotations():
                yield from orientation_changes(rotation, lambda x: x.rotate_clockwise())

    def _all_turns(self) -> Iterable[Tuple[Side, "Orientation"]]:
        yield Side.TOP, self.to_right
        yield Side.BOTTOM, self.to_left
        yield Side.RIGHT, self.to_bottom
        yield Side.LEFT, self.to_top

    @pipe(_normalize_turns)
    def turns_to(self, other: "Orientation") -> Iterable[Side]:
        orientation = self

        if orientation.front != other.front:
            for new_side, new_orientation in orientation._all_turns():
                if new_orientation.front == other.front:
                    yield new_side
                    orientation = new_orientation
                    break
            else:
                if orientation.to_bottom.front == other.front:
                    yield Side.RIGHT
                    yield Side.RIGHT
                    orientation = orientation.to_top.to_top
                else:
                    yield Side.TOP
                    yield Side.TOP
                    orientation = orientation.to_right.to_right

        while orientation.top != other.top:
            yield Side.FRONT
            orientation = orientation.rotate_clockwise()
        assert orientation == other



from typing import Tuple, Dict, List, Generic, TypeVar, Iterator, Optional
from .orientation import Side, Color, Orientation
from .sides import CubeSide, ICubeSide, CubeSideView
from .pattern import Pattern

T = TypeVar("T")


def shift_list(list: List[T], amount: int = 1) -> List[T]:
    return list[amount:] + list[:amount]


class Cube(Generic[T]):
    def __init__(self, shape: Tuple[int, int, int]):
        self.shape: Tuple[int, int, int] = shape
        self.sides: Dict[Side, CubeSide[T]] = {
            Side.FRONT: CubeSide[T](shape[0], shape[2], Color.RED),
            Side.LEFT: CubeSide[T](shape[1], shape[2], Color.BLUE),
            Side.BACK: CubeSide[T](shape[0], shape[2], Color.ORANGE),
            Side.RIGHT: CubeSide[T](shape[1], shape[2], Color.GREEN),
            Side.TOP: CubeSide[T](shape[0], shape[1], Color.YELLOW),
            Side.BOTTOM: CubeSide[T](shape[0], shape[1], Color.WHITE)
        }

    def get_side(self, orientation: Orientation) -> ICubeSide[T]:
        rotation = orientation.get_side_rotation()

        if rotation == 0:
            return self.sides[orientation.front]
        else:
            return CubeSideView(self.sides[orientation.front], rotation)

    @staticmethod
    def _fix_index(index: int, items_count: int) -> int:
        if index == 0:
            raise ValueError("index cannot be zero")
        elif abs(index) > items_count:
            raise ValueError(f"An absolute value of index is too large. "
                             f"{items_count} is an upper limit")
        elif index < 0:
            index = items_count + 1 + index
        return index - 1

    def turn_vertical(self, orientation: Orientation, index: int, turns: int) -> None:
        faces: List[ICubeSide] = []
        for i in range(4):
            faces.append(self.get_side(orientation))
            orientation = orientation.to_top

        index = self._fix_index(index, faces[0].columns)
        columns = [face.get_column(index) for face in faces]
        columns = shift_list(columns, 4 - turns)
        for face, column in zip(faces, columns):
            face.set_column(index, column)

        if index == 0:
            left_face = self.get_side(orientation.to_left)
            left_face.rotate(4 - turns)
        elif index == faces[0].columns - 1:
            right_face = self.get_side(orientation.to_right)
            right_face.rotate(turns)

    def turn_horizontal(self, orientation: Orientation, index: int, turns: int) -> None:
        orientation = orientation.rotate_counterclockwise()
        self.turn_vertical(orientation, index, turns)

    def turn_slice(self, orientation: Orientation, index: int, turns: int) -> None:
        orientation = orientation.to_right
        self.turn_vertical(orientation, index, 4 - turns)

    def get_data(self, orientation: Orientation, i: int, j: int) -> Optional[T]:
        return self.get_side(orientation)[i, j].data

    def set_data(self, orientation: Orientation, i: int, j: int, value: Optional[T]) -> None:
        front = self.get_side(orientation)
        front[i, j].data = value

        if j == 0:
            left = self.get_side(orientation.to_left)
            left[i, left.columns - 1].data = value
        elif j == front.columns - 1:
            right = self.get_side(orientation.to_right)
            right[i, 0].data = value

        if i == 0:
            top = self.get_side(orientation.to_top)
            top[top.rows - 1, j].data = value
        elif i == front.rows - 1:
            bottom = self.get_side(orientation.to_bottom)
            bottom[0, j].data = value

    def iterate_components(self) -> Iterator[Tuple[Side, int, int]]:
        for i in range(self.shape[2]):
            for j in range(self.shape[0]):
                yield Side.FRONT, i, j
                yield Side.BACK, i, j

        for j in range(1, self.shape[1] - 1):
            for i in range(1, self.shape[0] - 1):
                yield Side.TOP, j, i,
                yield Side.BOTTOM, j, i

            for k in range(self.shape[2]):
                yield Side.LEFT, k, j
                yield Side.RIGHT, k, j

    def get_absolute_coordinates(self, side: Side, i: int, j: int) -> Tuple[int, int, int]:
        if side == Side.FRONT:
            return j, i, 0
        elif side == Side.RIGHT:
            return self.shape[0] - 1, i, j
        elif side == Side.BACK:
            return self.shape[0] - 1 - j, i, self.shape[1] - 1
        elif side == Side.LEFT:
            return 0, i, self.shape[1] - 1 - j
        elif side == Side.TOP:
            return j, 0, self.shape[1] - 1 - i
        elif side == Side.BOTTOM:
            return j, self.shape[2] - 1, i
        raise ValueError("Unknown value for `side` argument")

    def orient(self, orientation: Orientation, keeping: Optional[Side] = None,
               front: Optional[Pattern] = None, top: Optional[Pattern] = None,
               left: Optional[Pattern] = None, right: Optional[Pattern] = None,
               back: Optional[Pattern] = None, bottom: Optional[Pattern] = None) -> Optional[Orientation]:
        def match(pattern: Optional[Pattern], front_orientation: Orientation, values: Dict) -> bool:
            if pattern is None:
                return True
            new_values = pattern.match(self.get_side(front_orientation), values)
            if new_values is None:
                return False
            for key in new_values:
                if key in values and values[key] != new_values[key]:
                    return False
            values.update(new_values)
            return True

        def perform_matching(new_front):
            values = {}
            yield match(front, new_front, values)
            yield match(top, new_front.to_top, values)
            yield match(left, new_front.to_left, values)
            yield match(right, new_front.to_right, values)
            yield match(bottom, new_front.to_bottom, values)
            yield match(back, new_front.to_left.to_left, values)

        for possible_front in orientation.iterate_rotations(keeping):
            if all(perform_matching(possible_front)):
                return possible_front
        else:
            return None



from types import MethodType
import typing


class Type:
    def __init__(self, name: str, lang_name: str, default: str):
        self.name = name
        self.lang_name = lang_name
        self.default = default

    def is_assignable(self, fr: "Type") -> bool:
        return self.get_generic_vars(fr) is not None

    def get_generic_vars(self, fr: "Type") -> typing.Optional[typing.Dict[str, "Type"]]:
        return {} if self == fr else None

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.lang_name

    def __repr__(self):
        return self.name

    def default_value(self) -> str:
        return self.default

    def substitute_generic(self, _generic_arguments: typing.Dict[str, "Type"]) -> "Type":
        return self


class GenericTypeVar(Type):
    def __init__(self, name: str):
        super().__init__(name, name, "")

    def get_generic_vars(self, fr: Type):
        return {self.name: fr}

    def substitute_generic(self, generic_arguments: typing.Dict[str, "Type"]):
        return generic_arguments[self.name]


T = GenericTypeVar("T")

Integer = Type("Integer", "int", "0")

Real = Type("Real", "real", "0.0")
Void = Type("Void", "void", "")
Real.is_assignable = MethodType(lambda self, fr: fr in {Integer, Real}, Real)
Color = Type("Color", "color", "white")
Side = Type("Side", "side", "front")
Bool = Type("Bool", "bool", "False")
Pattern = Type("Pattern", "pattern", "None")


class CollectionType(Type):
    def __init__(self, name: str, lang_name: str,
                 item_type: typing.Union[Type, GenericTypeVar], default: str):
        super().__init__(name, lang_name, default)
        self.item_type = item_type

    def __eq__(self, other):
        return type(self) == type(other) and self.item_type == other.item_type

    def __str__(self):
        return f"{self.lang_name} of {self.item_type}"

    def __repr__(self):
        return f"{self.name}({self.item_type!r})"

    def __hash__(self):
        return hash((self.name, hash(self.item_type)))

    def get_generic_vars(self, fr: "Type"):
        if type(self) != type(fr):
            return None
        fr: CollectionType
        return self.item_type.get_generic_vars(fr.item_type)


class List(CollectionType):
    def __init__(self, item_type: Type):
        super().__init__("List", "list", item_type, "list()")

    def substitute_generic(self, generic_arguments: typing.Dict[str, "Type"]):
        return List(self.item_type.substitute_generic(generic_arguments))


class Set(CollectionType):
    def __init__(self, item_type: Type):
        super().__init__("Set", "set", item_type, "set()")

    def substitute_generic(self, generic_arguments: typing.Dict[str, "Type"]):
        return Set(self.item_type.substitute_generic(generic_arguments))


def type_annotation_to_type(annotation: typing.Any) -> Type:
    if annotation == int:
        return Integer
    elif annotation == float:
        return Real
    elif annotation == bool:
        return Bool
    elif annotation is None:
        return Void
    r = repr(annotation)
    if r.startswith(repr(typing.List)):
        return List(type_annotation_to_type(annotation.__args__[0]))
    elif r.startswith(repr(typing.Set)):
        return Set(type_annotation_to_type(annotation.__args__[0]))
    else:
        raise ValueError(f"Unsupported annotation: {annotation!r}")


class Function(Type):
    class FunctionOverload(typing.NamedTuple):
        arguments: typing.List[Type]
        return_type: Type

    def __init__(self, *overloads: typing.Tuple[typing.List[Type], Type]):
        super().__init__("Function", "func", "")
        self.overloads = [Function.FunctionOverload(*x) for x in overloads]

    def prepend_overload(self, arguments: typing.List[Type], return_type: Type):
        self.overloads.insert(0, self.FunctionOverload(arguments, return_type))

    def __hash__(self):
        return hash(tuple((tuple(x.arguments), x.return_type) for x in self.overloads))

    def __eq__(self, other):
        return isinstance(other, Function) and self.overloads == other.overloads

    def __repr__(self):
        args = ", ".join(f"({x.arguments}, {x.return_type})" for x in self.overloads)
        return f"{self.name}({args})"

    def expand_variadic(types: typing.List[typing.Union[Type, type(Ellipsis)]], 
                        number: int) -> typing.Optional[typing.List[Type]]:
        result: List[Type] = []
        previous = None
        for t in types:
            if t == Ellipsis:
                if previous == None:
                    raise ValueError("Ellipsis cannot be the first argument of the function")
                while len(result) < number:
                    result.append(previous)
                return result
            else:
                result.append(t)
                if len(result) > number:
                    return None
                previous = t
        return result if len(result) == number else None

    def takes_arguments(self, arguments: typing.List[Type]) -> typing.Optional[Type]:
        for overload in self.overloads:
            func_arguments = Function.expand_variadic(overload.arguments, len(arguments))
            if func_arguments is None:
                continue
            generic_arguments = {}
            for func_arg, real_arg in zip(func_arguments, arguments):
                if func_arg == Real and real_arg == Integer:
                    continue
                else:
                    generics = func_arg.get_generic_vars(real_arg)
                    if generics is None:
                        break
                    elif any(argument in generic_arguments and generics[argument] != generic_arguments[argument]
                             for argument in generics):
                        break
                    generic_arguments.update(generics)
            else:
                return overload.return_type.substitute_generic(generic_arguments)
        return None



import enum
from abc import ABC, abstractmethod
from typing import Union, List, Iterable, TypeVar, Optional, Set, Dict

from .cube import Cube
from .orientation import Orientation, Side, count_occurrences

T = TypeVar("T")


class Action(ABC):
    @abstractmethod
    def perform(self, cube: Cube, orientation: Orientation) -> Orientation:
        pass


class Rotate(Action):
    _ROTATION_LETTERS: Dict[Side, str] = {
        Side.RIGHT: "X",
        Side.TOP: "Y",
        Side.FRONT: "Z"
    }

    def __init__(self, around: Side, twice: bool = False) -> None:
        self.axis_side: Side = around
        self.twice = twice

    def perform_single(self, orientation: Orientation) -> Orientation:
        if self.axis_side == Side.FRONT:
            return orientation.rotate_clockwise()
        elif self.axis_side == Side.BACK:
            return orientation.rotate_counterclockwise()
        elif self.axis_side == Side.RIGHT:
            return orientation.to_bottom
        elif self.axis_side == Side.LEFT:
            return orientation.to_top
        elif self.axis_side == Side.TOP:
            return orientation.to_right
        else:
            return orientation.to_left

    def perform(self, cube: Optional[Cube], orientation: Orientation) -> Orientation:
        if self.twice:
            return self.perform_single(self.perform_single(orientation))
        else:
            return self.perform_single(orientation)

    def __repr__(self):
        return f"Rotate({self.axis_side}, {self.twice})"

    def __str__(self):
        if self.axis_side in Rotate._ROTATION_LETTERS:
            letter = Rotate._ROTATION_LETTERS[self.axis_side]
        else:
            letter = Rotate._ROTATION_LETTERS[self.axis_side.opposite()]
            if not self.twice:
                letter += "'"
        if self.twice:
            return letter + "2"
        else:
            return letter

    @staticmethod
    def from_turn_steps(steps: Iterable[Side]) -> Iterable["Rotate"]:
        for side, turns in count_occurrences(steps):
            turns = turns % 4
            if turns == 3:
                side = side.opposite()
            if turns != 0:
                yield Rotate(side, turns == 2)


class TurningType(enum.Enum):
    HORIZONTAL = enum.auto()
    VERTICAL = enum.auto()
    SLICE = enum.auto()


class Turn(Action):
    TYPES = {
        Side.LEFT: TurningType.VERTICAL, Side.RIGHT: TurningType.VERTICAL,
        Side.TOP: TurningType.HORIZONTAL, Side.BOTTOM: TurningType.HORIZONTAL,
        Side.FRONT: TurningType.SLICE, Side.BACK: TurningType.SLICE
    }

    def __init__(self, side: Union[Side, TurningType],
                 indices: Union[int, List[Union[int, type(Ellipsis)]]],
                 turns: int = 1) -> None:
        self.indices: List[Union[int, type(Ellipsis)]] = \
            indices if isinstance(indices, list) else [indices]
        if not (all(x == Ellipsis or x > 0 for x in self.indices) or
                all(x == Ellipsis or x < 0 for x in self.indices)):
            raise ValueError("All indices of the turn action must be either "
                             "positive or negative")
        self.turns: int = turns

        self.type: TurningType
        if isinstance(side, Side):
            self.type = Turn.TYPES[side]

            if side in {Side.BACK, Side.RIGHT, Side.BOTTOM}:
                self.indices = Turn.opposite_side(self.indices)

            if side not in {Side.BOTTOM, Side.RIGHT, Side.FRONT}:
                self.turns: int = 4 - turns
        else:
            self.type = side

    def perform(self, cube: Cube, orientation: Orientation) -> Orientation:
        turning_functions = {
            TurningType.VERTICAL: (cube.turn_vertical, cube.get_side(orientation).columns),
            TurningType.HORIZONTAL: (cube.turn_horizontal, cube.get_side(orientation).rows),
            TurningType.SLICE: (cube.turn_slice, cube.get_side(orientation.to_right).columns)
        }
        rotate_function, size = turning_functions[self.type]
        for side in Turn.normalize_indices(self.indices, size):
            rotate_function(orientation, side, self.turns)
        return orientation

    def __repr__(self):
        return f"Turn({self.type}, {self.indices}, {self.turns})"

    def __str__(self):
        opposite = any(x != Ellipsis and x < 0 for x in self.indices)
        if self.type == TurningType.VERTICAL:
            letter = "L" if not opposite else "R"
        elif self.type == TurningType.HORIZONTAL:
            letter = "U" if not opposite else "D"
        else:
            letter = "F" if not opposite else "B"

        if self.turns == 2:
            letter += "2"
        else:
            turns = self.turns
            if letter not in {"D", "R", "F"}:
                turns = 4 - self.turns
            if turns == 3:
                letter += "'"

        if self.indices != [1] and self.indices != [-1]:
            string = []
            prev_ellipsis = True
            for index in self.indices:
                if index == Ellipsis:
                    string.append(":")
                    prev_ellipsis = True
                else:
                    if not prev_ellipsis:
                        string.append(",")
                    prev_ellipsis = False
                    string.append(str(abs(index)))
            letter += "[" + "".join(string) + "]"
        return letter

    def _transform(self, turn: Side) -> "Turn":
        if Turn.TYPES[turn] == self.type:
            return self

        if turn == Side.TOP:
            if self.type == TurningType.SLICE:
                return Turn(TurningType.VERTICAL, self.indices, 4 - self.turns)
            else:  # TurningType.VERTICAL
                return Turn(TurningType.SLICE, Turn.opposite_side(self.indices), self.turns)
        elif turn == Side.FRONT:
            if self.type == TurningType.VERTICAL:
                return Turn(TurningType.HORIZONTAL, self.indices, self.turns)
            else:  # TurningType.HORIZONTAL
                return Turn(TurningType.VERTICAL, Turn.opposite_side(self.indices), 4 - self.turns)
        elif turn == Side.RIGHT:
            if self.type == TurningType.SLICE:
                return Turn(TurningType.HORIZONTAL, self.indices, 4 - self.turns)
            else:  # TurningType.HORIZONTAL
                return Turn(TurningType.SLICE, Turn.opposite_side(self.indices), self.turns)
        else:
            raise ValueError("Unsupported turn")

    def from_orientation(self, orientation: Orientation, origin=Orientation()) -> "Turn":
        result: Turn = self
        for turn in list(orientation.turns_to(origin)):
            result = result._transform(turn)
        return result

    @staticmethod
    def opposite_side(indices: List[Union[int, type(Ellipsis)]]) \
            -> List[Union[int, type(Ellipsis)]]:
        return [Ellipsis if index == Ellipsis else -index for index in indices]

    @staticmethod
    def normalize_indices(indices: List[Union[int, type(Ellipsis)]], width: int) -> Set[int]:
        def to_positive(idx: Union[int, type(Ellipsis)]):
            if idx == Ellipsis:
                return ...
            elif idx < 0:
                return width + idx + 1
            else:
                return idx

        def add_ends(it: Iterable[Union[int, type(Ellipsis)]]):
            yield 0
            yield from it
            yield width + 1

        def get_tripples(it: Iterable[Union[int, type(Ellipsis)]]):
            v_iter = iter(it)
            a = next(v_iter)
            b = next(v_iter)
            while True:
                try:
                    c = next(v_iter)
                    yield a, b, c
                    a, b = b, c
                except StopIteration:
                    break

        result = set()
        stream = add_ends(map(to_positive, indices))
        for prev_value, value, next_value in get_tripples(stream):
            if value == Ellipsis:
                if prev_value > next_value:
                    prev_value, next_value = next_value, prev_value
                result.update(range(prev_value + 1, next_value))
            else:
                result.add(value)
        return result



from typing import Dict, NamedTuple, Optional

from .expression import VariablesPool
from .types import Type


class VariableDefinition(NamedTuple):
    type: Type
    number: int


class StackFrame:
    def __init__(self, top_frame: Optional["StackFrame"]):
        self.variables: Dict[str, VariableDefinition] = dict()
        self.top_frame: Optional[StackFrame] = top_frame


class Stack:
    def __init__(self, return_type: Optional[Type] = None):
        self.globals: Dict[str, VariableDefinition] = {}
        self.stack_top: StackFrame = StackFrame(None)
        self.pool: VariablesPool = VariablesPool()
        self.context_return_type: Optional[Type] = return_type

    def add_frame(self) -> None:
        self.stack_top = StackFrame(self.stack_top)

    def pop_frame(self) -> None:
        self.pool.deallocate(len(self.stack_top.variables))
        self.stack_top = self.stack_top.top_frame

    def get_variable(self, name: str, ) -> Optional[VariableDefinition]:
        frame = self.stack_top
        while frame is not None:
            if name in frame.variables:
                return frame.variables[name]
            frame = frame.top_frame
        if name in self.globals:
            return self.globals[name]
        return None

    def add_variable(self, name: str, var_type: Type) -> int:
        number = self.pool.allocate_single()
        definition = VariableDefinition(var_type, number)
        self.stack_top.variables[name] = definition
        return number

    def add_global(self, name: str, var_type: Type) -> None:
        self.globals[name] = VariableDefinition(var_type, -1)

    def create_inner(self, return_type: Optional[Type]) -> "Stack":
        stack = Stack(return_type)
        stack.globals = self.globals
        return stack



from typing import List, Optional, Union, Tuple

from .code_map import CodeMap
from .codeio import CodeStream
from .types import Type, Void, Integer

TemplateType = List[Union[str, int]]


class VariablesPool:
    class Context:
        def __init__(self, pool: "VariablesPool", allocated: int):
            self.pool = pool
            self._allocated: int = allocated

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.pool.deallocate(self._allocated)
            return False

        def __iter__(self):
            return range(self.pool._allocated - self._allocated, self.pool._allocated).__iter__()

    def __init__(self):
        self._allocated = 0

    def allocate(self, count: int) -> Context:
        self._allocated += count
        return VariablesPool.Context(self, count)

    def allocate_single(self):
        self._allocated += 1
        return self._allocated - 1

    def deallocate(self, count: int):
        self._allocated -= count


class Expression:
    def __init__(self, line_number: int, expr_type: Type, text: Union[str, TemplateType] = None):
        self.line_number: int = line_number
        self.intermediates: List[Expression] = list()
        self.type: Type = expr_type
        if text is not None and isinstance(text, str):
            text = [text]
        self.expression: TemplateType = text

    def add_intermediate(self, expression: "Expression") -> int:
        self.intermediates.append(expression)
        return len(self.intermediates) - 1

    def generate_intermediates(self, vars: VariablesPool.Context, pool: VariablesPool, stream: CodeStream,
                               code_map: CodeMap):
        for var, expr in zip(vars, self.intermediates):
            temp_var_name = "tmp_" + str(var)
            expr.generate(pool, stream, code_map, temp_var_name)

    def generate_expression_line(self, vars: VariablesPool.Context) -> str:
        variables = list(vars)
        return "".join("tmp_" + str(variables[c]) if isinstance(c, int) else c
                       for c in self.expression)

    def generate_line(self, pool: VariablesPool, stream: CodeStream, code_map: CodeMap) -> str:
        with pool.allocate(len(self.intermediates)) as variables:
            self.generate_intermediates(variables, pool, stream, code_map)
            return self.generate_expression_line(variables)

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        if self.expression == [0]:
            self.intermediates[0].generate(temp_pool, stream, code_map, var_name)
            return
        expression = self.generate_line(temp_pool, stream, code_map)
        if var_name is not None:
            expression = var_name + " = " + expression
        code_map.add(stream.line_number, self.line_number)
        stream.push_line(expression)

    @staticmethod
    def _from_template(template: TemplateType, expressions: List[TemplateType]) -> TemplateType:
        for component in template:
            if isinstance(component, int):
                yield from expressions[component]
            else:
                yield component

    @staticmethod
    def merge(expr_type: Type, template: TemplateType, *parts: "Expression") -> "Expression":
        line_number = None if any(x is None for x in parts) else min(x.line_number for x in parts)
        result = Expression(line_number, expr_type)

        expressions = [parts[0].expression]
        for inter in parts[0].intermediates:
            result.add_intermediate(inter)
        offset = len(parts[0].intermediates)

        for expr in parts[1:]:
            this_expression = expr.expression[::]
            for i, component in enumerate(this_expression):
                if isinstance(component, int):
                    this_expression[i] += offset
            expressions.append(this_expression)
            for inter in expr.intermediates:
                result.add_intermediate(inter)
            offset += len(expr.intermediates)

        result.expression = list(Expression._from_template(template, expressions))
        return result


class ConditionExpression(Expression):
    class Intermediate(Expression):
        def __init__(self, line_number: int, returns_value: bool, actions: List[Tuple[Expression, List[Expression]]],
                     else_clause: List[Expression]):
            super(ConditionExpression.Intermediate, self).__init__(line_number, Void)
            self.returns_value = returns_value
            self.actions: List[Tuple[Expression, List[Expression]]] = actions
            self.else_clause: List[Expression] = else_clause

        @staticmethod
        def _generate_clause(clause: List[Expression], temp_pool: VariablesPool,
                             stream: CodeStream, code_map: CodeMap, var_name: Optional[str]):
            for line in clause[:-1]:
                line.generate(temp_pool, stream, code_map, None)
            clause[-1].generate(temp_pool, stream, code_map, var_name)

        def _generate_if(self, actions: List[Tuple[Expression, List[Expression]]], temp_pool: VariablesPool,
                         stream: CodeStream, code_map: CodeMap, var_name: Optional[str]):
            condition = actions[0][0].generate_line(temp_pool, stream, code_map)
            code_map.add(stream.line_number, actions[0][0].line_number)
            stream.push_line(f"if {condition}:")
            stream.indent()
            self._generate_clause(actions[0][1], temp_pool, stream, code_map, var_name)
            stream.unindent()
            if len(actions) > 1 or len(self.else_clause) > 0:
                stream.push_line("else:")
                stream.indent()
                if len(actions) > 1:
                    self._generate_if(actions[1:], temp_pool, stream, code_map, var_name)
                else:
                    self._generate_clause(self.else_clause, temp_pool, stream, code_map, var_name)
                stream.unindent()

        def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap,
                     var_name: Optional[str] = None):
            if not self.returns_value:
                var_name = None
            self._generate_if(self.actions, temp_pool, stream, code_map, var_name)

    def __init__(self, line_number: int, actions: List[Tuple[Expression, List[Expression]]],
                 else_clause: List[Expression]):
        if len(else_clause) > 0:
            res_type = else_clause[-1].type
            for action in actions:
                action_type = action[1][-1].type
                if action_type.is_assignable(res_type):
                    res_type = action_type
                elif not res_type.is_assignable(action_type):
                    res_type = Void
                    break
        else:
            res_type = Void

        super(ConditionExpression, self).__init__(line_number, res_type, [0])
        self.add_intermediate(ConditionExpression.Intermediate(line_number, res_type, actions, else_clause))


class WhileLoopExpression(Expression):
    def __init__(self, line_number: int, condition: Expression, actions: List[Expression]):
        super().__init__(line_number, Void)
        self.condition: Expression = condition
        self.actions: List[Expression] = actions

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        with temp_pool.allocate(len(self.condition.intermediates)) as cond_vars:
            self.condition.generate_intermediates(cond_vars, temp_pool, stream, code_map)
            code_map.add(stream.line_number, self.condition.line_number)
            stream.push_line("while " + self.condition.generate_expression_line(cond_vars) + ":")
            stream.indent()
            for action in self.actions:
                action.generate(temp_pool, stream, code_map, None)
            self.condition.generate_intermediates(cond_vars, temp_pool, stream, code_map)
            stream.unindent()


class RepeatLoopExpression(Expression):
    def __init__(self, line_number: int, times: Expression, actions: List[Expression]):
        super().__init__(line_number, Void)
        self.times: Expression = times
        self.actions: List[Expression] = actions

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        with temp_pool.allocate(1) as ctr:
            counter, = ctr
            line = self.times.generate_line(temp_pool, stream, code_map)
            code_map.add(stream.line_number, self.times.line_number)
            stream.push_line(f"for tmp_{counter} in range({line}):")
            stream.indent()
            for action in self.actions:
                action.generate(temp_pool, stream, code_map, None)
            stream.unindent()


class DoWhileLoopExpression(Expression):
    def __init__(self, line_number: int, condition: Expression, actions: List[Expression]):
        super().__init__(line_number, Void)
        self.condition = condition
        self.actions = actions

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        stream.push_line("while True:")
        stream.indent()
        for action in self.actions:
            action.generate(temp_pool, stream, code_map, None)

        condition = self.condition.generate_line(temp_pool, stream, code_map)
        code_map.add(stream.line_number, self.condition.line_number)
        stream.push(f"if not ({condition}):")
        stream.indent()
        stream.push("break")
        stream.unindent(2)


class ForLoopExpression(Expression):
    def __init__(self, line_number: int, iterator: str, loop_range: Expression, actions: List[Expression]):
        super().__init__(line_number, Void)
        self.iterator: str = iterator
        self.range: Expression = loop_range
        self.actions: List[Expression] = actions

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        range_expression = self.range.generate_line(temp_pool, stream, code_map)
        code_map.add(stream.line_number, self.range.line_number)
        stream.push(f"for {self.iterator} in {range_expression}:")
        stream.indent()
        for action in self.actions:
            action.generate(temp_pool, stream, code_map, None)
        stream.unindent()


class CubeTurningExpression(Expression):
    def __init__(self, line_number: int, side: str, amount: int):
        super().__init__(line_number, Void)
        self.side: str = side
        self.amount: int = amount
        self.indices: List[Union[Expression, type(Ellipsis)]] = [Expression(line_number, Integer, "1")]

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap,
                 var_name: Optional[str] = None):
        code_map.add(stream.line_number, self.line_number)
        expression = [f"cube_turn({self.side}, {self.amount}, ["]
        for i in range(len(self.indices)):
            expression.append(i)
            expression.append(", ")
        expression[-1] = "])"

        Expression.merge(self.type, expression, *self.indices).generate(temp_pool, stream, code_map, var_name)


class CubeRotationExpression(Expression):
    def __init__(self, line_number: int, side: str, twice: bool = False):
        super().__init__(line_number, Void)
        self.side: str = side
        self.twice: bool = twice

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        code_map.add(stream.line_number, self.line_number)
        stream.push_line(f"cube_rotate({self.side}, {self.twice})")


class FunctionDeclarationExpression(Expression):
    def __init__(self, line_number: int, name: str, symbol_name: str,
                 return_type: Type, arguments: List[str], clause: List[Expression]):
        super().__init__(line_number, Void)
        self.name: str = name
        self.symbol_name: str = symbol_name
        self.return_type: Type = return_type
        self.arguments: List[str] = arguments
        self.clause: List[Expression] = clause

    def generate(self, temp_pool: VariablesPool, stream: CodeStream, code_map: CodeMap, var_name: Optional[str] = None):
        arguments = ", ".join(self.arguments)
        stream.push_line(f"@runtime_function(\"{self.symbol_name}\")")
        stream.push_line(f"def {self.name}({arguments}):")
        stream.indent()
        for expression in self.clause:
            expression.generate(temp_pool, stream, code_map, None)
        if self.return_type != Void:
            stream.push_line("return " + self.return_type.default_value())
        elif len(self.clause) == 0:
            stream.push_line("pass")
        stream.unindent()



from typing import Dict


class CodeMap:
    def __init__(self):
        self.line_numbers: Dict[int, int] = dict()

    def add(self, python_line: int, source_line: int) -> None:
        if source_line is not None and python_line is not None:
            self.line_numbers[python_line] = source_line

    def __getitem__(self, index: int):
        while index > 0 and index not in self.line_numbers:
            index -= 1
        if index in self.line_numbers:
            return self.line_numbers[index]
        else:
            return 0



import random
from argparse import ArgumentParser
from typing import List

from cubelang.actions import Turn
from cubelang.cli.options import integer_type
from cubelang.cube import Cube
from cubelang.orientation import Orientation, Side

# noinspection PyTypeChecker
SIDES = tuple(Side)


def main():
    arg_parser = ArgumentParser()
    arg_parser.add_argument("-d", dest="dimension", help="dimensions of a cube",
                            default=3, metavar="N", type=integer_type(2))
    arg_parser.add_argument("-n", dest="turns_num", help="number of turns",
                            type=integer_type(1), default=20)
    arg_parser.add_argument("-a", dest="output_args", action="store_true",
                            help="display the state of the cube after the turns instead of the formula")
    arg_parser.add_argument("-s", dest="seed", help="the seed for the pseudorandom number generator")
    args = arg_parser.parse_args()
    dim = args.dimension

    if args.seed is not None:
        random.seed(args.seed)

    actions: List[Turn] = []
    prev_side = None
    for i in range(args.turns_num):
        if prev_side is None:
            sides = SIDES
        else:
            sides = [x for x in SIDES if x != prev_side]

        prev_side = random.choice(sides)

        first_index = random.randint(1, dim // 2)
        last_index = random.randint(1, first_index)
        if first_index == last_index:
            indices = [first_index]
        else:
            indices = [last_index, ..., first_index]

        turn = Turn(prev_side, indices, random.randint(1, 3))
        actions.append(turn)

    if not args.output_args:
        for action in actions:
            print(str(action), end="")
        print()
    else:
        cube = Cube((dim,) * 3)
        orientation = Orientation()
        for action in actions:
            action.perform(cube, orientation)

        print("--front", repr(cube.get_side(orientation).colors))
        print("--right", repr(cube.get_side(orientation.to_right).colors))
        print("--left", repr(cube.get_side(orientation.to_left).colors))
        print("--back", repr(cube.get_side(orientation.to_right.to_right).colors))
        print("--top", repr(cube.get_side(orientation.to_top).colors))
        print("--bottom", repr(cube.get_side(orientation.to_bottom).colors))


if __name__ == "__main__":
    main()



from argparse import ArgumentParser, Namespace, ArgumentTypeError

from .options import formula_type, side_colors_type, integer_type
from ..actions import Action
from ..cube import Cube
from ..orientation import Orientation, Color, Side
from ..cube_runtime import CubeRuntime
from typing import Optional, List, Tuple


def apply_side(cube: Cube, orientation: Orientation,
               colors: Optional[List[List[Color]]]):
    if colors is None:
        return
    side = cube.get_side(orientation)
    if side.rows != len(colors):
        raise ArgumentTypeError(f"Incorrect number of lines")
    elif side.columns != len(colors[0]):
        raise ArgumentTypeError(f"Incorrect number of columns")

    for i, line in enumerate(colors):
        for j, color in enumerate(line):
            side.colors[i, j] = color


class CubeBuilder:
    def __init__(self, size: Tuple[int, int, int]):
        self.cube = Cube(size)
        self.orientation = Orientation()

    def scramble(self, actions: List[Action]) -> "CubeBuilder":
        for action in actions:
            self.orientation = action.perform(self.cube, self.orientation)
        return self

    def side(self, side: Side, colors: List[List[Color]]) -> "CubeBuilder":
        orientation = Orientation.regular(side)
        apply_side(self.cube, orientation, colors)
        return self

    def get(self) -> Tuple[Cube, Orientation]:
        return self.cube, self.orientation


def init_cube_args_parser(argparse: ArgumentParser):
    group = argparse.add_argument_group("cube options")
    group.add_argument("-d", dest="dimension", help="dimensions of a cube",
                       default=3, metavar="N", type=integer_type(2))
    group.add_argument("-s", dest="scramble", help="formula to scramble a cube",
                       default=[], type=formula_type, metavar="FORMULA")
    for name, side in CubeRuntime.SIDE_NAMES.items():
        group.add_argument("--" + name, type=side_colors_type, metavar="COLORS",
                           help=f"colors of the {name} face of a cube", default=None)


def build_cube(arguments: Namespace):
    builder = CubeBuilder((arguments.dimension,) * 3)
    for name, side in CubeRuntime.SIDE_NAMES.items():
        builder.side(side, getattr(arguments, name))
    return builder.scramble(arguments.scramble).get()



from typing import Callable
from unittest.mock import MagicMock
from unittest.mock import patch

import pytest

from cubelang.actions import Action, Turn, TurningType
from cubelang.cube import Cube
from cubelang.cube_runtime import CubeRuntime
from cubelang.orientation import Side, Orientation


def test_runtime_globals():
    runtime = CubeRuntime(Cube((3, 3, 3)), Orientation(), lambda action: None, lambda: None)
    existing = set(runtime.functions.global_values.keys())
    assert existing.issuperset(set(CubeRuntime.COLOR_NAMES.keys()))
    assert existing.issuperset(set(CubeRuntime.SIDE_NAMES.keys()))
    assert "push_orientation" in existing


def test_action_callback():
    callback: Callable[[Action], None] = MagicMock()
    runtime = CubeRuntime(Cube((3, 3, 3)), Orientation(), callback, lambda: None)
    runtime.perform_turn(Side.LEFT, 2, [1])

    callback: MagicMock
    action = callback.call_args_list[0][0][0]
    assert isinstance(action, Turn)
    assert action.type == TurningType.VERTICAL


@pytest.mark.parametrize("side, orientation", [
    (Side.FRONT, Orientation(Side.FRONT, Side.TOP)),
    (Side.LEFT, Orientation(Side.LEFT, Side.TOP)),
    (Side.RIGHT, Orientation(Side.RIGHT, Side.TOP)),
    (Side.BACK, Orientation(Side.BACK, Side.TOP)),
    (Side.TOP, Orientation(Side.TOP, Side.BACK)),
    (Side.BOTTOM, Orientation(Side.BOTTOM, Side.FRONT))
])
def test_get_color(side, orientation):
    class Colors:
        def __getitem__(self, item):
            pass

    class Side:
        def __init__(self):
            self.colors = Colors()

    with patch.object(Cube, 'get_side', return_value=Side()) as mock_method:
        runtime = CubeRuntime(Cube((3, 3, 3)), Orientation(), lambda action: None, lambda: None)
        runtime.get_color(side, 0, 0)
        mock_method.assert_called_once_with(orientation)


def test_state_stack():
    actions = []
    runtime = CubeRuntime(Cube((2, 2, 2)), Orientation(), actions.append, lambda: None)

    runtime.perform_turn(Side.FRONT, 1, [1])
    runtime.perform_rotate(Side.TOP, False)
    runtime.push_orientation()
    for _ in range(3):
        runtime.perform_turn(Side.FRONT, 1, [1])
        runtime.perform_rotate(Side.TOP, False)
    runtime.pop_orientation()

    assert "FYFYFYFYY" == "".join(map(str, actions))


def test_suspend_rotations():
    actions = []
    runtime = CubeRuntime(Cube((2, 2, 2)), Orientation(), actions.append, lambda: None)
    runtime.perform_turn(Side.FRONT, 1, [1])
    runtime.perform_rotate(Side.TOP, False)
    runtime.suspend_rotations()
    for _ in range(3):
        runtime.perform_turn(Side.FRONT, 1, [1])
        runtime.perform_rotate(Side.TOP, False)
    runtime.resume_rotations()

    assert "FYFRBY'" == "".join(map(str, actions))



import lark
from .types import Type, Function
from .expression import Expression

from typing import Optional, Union, List


class CompileTimeError(Exception):
    def __init__(self, node: lark.Tree, message: str):
        super(CompileTimeError, self).__init__(message)
        if isinstance(node, lark.Token) or hasattr(node.meta, "line"):
            self.start_line = node.line
            self.start_column = node.column
            self.end_line = node.end_line
            self.end_column = node.end_column


class ValueTypeError(CompileTimeError):
    def __init__(self, node: lark.Tree, message: str, expected: Optional[Type], actual: Type):
        super(ValueTypeError, self).__init__(node, message)
        self.expected = expected
        self.actual = actual


class UnresolvedReferenceError(CompileTimeError):
    def __init__(self, node: lark.Tree, name: Optional[str] = None):
        if name is None:
            name = str(node)
        message = f"Unresolved symbol: `{name}`"
        super(UnresolvedReferenceError, self).__init__(node, message)


class FunctionArgumentsError(CompileTimeError):
    def __init__(self, node: lark.Tree, function_name: str,
                 function: Function, arguments: List[Type]):
        super(FunctionArgumentsError, self).__init__(node, "Invalid function arguments")
        self.function_name = function_name
        self.function: Function = function
        self.arguments = arguments


def assert_type(node: lark.Tree, expression: Union[Expression, Type],
                required: Type, message: Optional[str] = None):
    val_type = expression if isinstance(expression, Type) else expression.type
    if required.is_assignable(val_type):
        return
    if message is None:
        message = f"Expected value of type {required} but a {val_type} was found."
    raise ValueTypeError(node, message, required, val_type)



import pytest

from cubelang.actions import Turn, Action, Rotate, TurningType
from cubelang.orientation import Side
from cubelang.parser import ParsingError, parse_actions


def test_valid_parsing():
    actions = list(parse_actions("R  L2  U' RU L"))
    expected_sides = [TurningType.VERTICAL, TurningType.VERTICAL,
                      TurningType.HORIZONTAL, TurningType.VERTICAL,
                      TurningType.HORIZONTAL, TurningType.VERTICAL]
    expected_turns = [1, 2, 1, 1, 3, 3]

    assert len(actions) == len(expected_sides)
    for action, type, turns in zip(actions, expected_sides, expected_turns):
        assert isinstance(action, Turn)
        assert action.type == type
        assert action.turns == turns


@pytest.mark.parametrize("text, type, indices, turns", [
    ("F", TurningType.SLICE, [1], 1),
    ("F'", TurningType.SLICE, [1], 3),
    ("B", TurningType.SLICE, [-1], 3),
    ("B'", TurningType.SLICE, [-1], 1),

    ("R", TurningType.VERTICAL, [-1], 1),
    ("R'", TurningType.VERTICAL, [-1], 3),
    ("L", TurningType.VERTICAL, [1], 3),
    ("L'", TurningType.VERTICAL, [1], 1),

    ("U", TurningType.HORIZONTAL, [1], 3),
    ("U'", TurningType.HORIZONTAL, [1], 1),
    ("D", TurningType.HORIZONTAL, [-1], 1),
    ("D'", TurningType.HORIZONTAL, [-1], 3),

    ("L[2]", TurningType.VERTICAL, [2], 3),
    ("L[:2]", TurningType.VERTICAL, [..., 2], 3),
    ("L2[1:2]", TurningType.VERTICAL, [1, ..., 2], 2),
    ("L[2:]", TurningType.VERTICAL, [2, ...], 3),
    ("L[1,3]", TurningType.VERTICAL, [1, 3], 3),
    ("L'[1:2,3]", TurningType.VERTICAL, [1, ..., 2, 3], 1),
    ("L[1,2:3]", TurningType.VERTICAL, [1, 2, ..., 3], 3)
])
def test_turn_single(text: str, type: TurningType, indices: int, turns: int):
    actions = list(parse_actions(text))
    assert len(actions) == 1
    action = actions[0]
    assert isinstance(action, Turn)
    assert action.indices == indices
    assert action.turns == turns
    assert action.type == type

    assert text == str(action)


def test_rotation():
    actions = list(parse_actions("X X' Y' Z2"))
    expected_sides = [Side.RIGHT, Side.LEFT, Side.BOTTOM, Side.FRONT]
    expected_twice = [False, False, False, True]

    assert len(actions) == len(expected_sides)
    for action, side, twice in zip(actions, expected_sides, expected_twice):
        assert isinstance(action, Rotate)
        assert action.twice == twice
        assert action.axis_side == side


@pytest.mark.parametrize("action, expected", [
    (Turn(Side.FRONT, 1, 1), "F"),
    (Turn(Side.FRONT, 1, 2), "F2"),
    (Turn(Side.FRONT, 1, 3), "F'"),
    (Turn(Side.LEFT, 1, 1), "L"),
    (Turn(Side.RIGHT, 1, 2), "R2"),
    (Turn(Side.BACK, 1, 3), "B'"),
    (Turn(Side.TOP, 1, 1), "U"),
    (Turn(Side.BOTTOM, 1, 2), "D2"),

    (Rotate(Side.FRONT, False), "Z"),
    (Rotate(Side.BACK, False), "Z'"),
    (Rotate(Side.BACK, True), "Z2"),
    (Rotate(Side.FRONT, True), "Z2"),
    (Rotate(Side.RIGHT, False), "X"),
    (Rotate(Side.BOTTOM, False), "Y'")
])
def test_representation(action: Action, expected: str):
    actual = str(action)
    assert expected == actual


@pytest.mark.parametrize("text, column", [
    ("Q", 0), ("X[1]", 1), ("R[", 2), ("R2'", 2), ("R'2", 2), ("R:", 1), ("R[]", 2),
    ("R[1,]", 4), ("R[,1]", 2), ("R[1,:]", 4), ("R[1:, 1]", 4)
])
def test_illegal_parsing(text: str, column: int):
    with pytest.raises(ParsingError) as e:
        parse_actions(text)
    assert e.value.column == column



from io import StringIO


class CodeStream:
    INDENTATION = "    "

    def __init__(self):
        self.io = StringIO()
        self.indentation: int = 0
        self.line_number: int = 0

    def push_line(self, text: str) -> None:
        self.io.write((CodeStream.INDENTATION * self.indentation) + text + "\n")
        self.line_number += 1

    def push(self, text: str) -> None:
        for line in text.split("\n"):
            self.push_line(line)

    def indent(self, by: int = 1) -> None:
        self.indentation += by

    def unindent(self, by: int = 1) -> None:
        self.indentation -= by

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_tb is None:
            self.io.close()

    def get_contents(self):
        self.io.seek(0)
        return self.io.read()



from typing import Optional

from ..actions import Action, Turn, Rotate
from .base import PostprocessorBase
from ..orientation import Orientation


class OrientationFreezePostprocessor(PostprocessorBase[Action, Turn]):
    def __init__(self, origin: Optional[Orientation] = None):
        super().__init__()
        if origin is not None:
            self.origin = origin
        else:
            self.origin = Orientation()
        self.current = self.origin

    def process(self, action: Action):
        if isinstance(action, Rotate):
            self.current = action.perform(None, self.current)
        else:
            assert isinstance(action, Turn)
            action = action.from_orientation(self.current, self.origin)
            self._return(action)



import traceback
from typing import Optional, NamedTuple, List


class TerminateExecutionError(Exception):
    pass


class RuntimeError(Exception):
    ACCEPTED_EXCEPTIONS = [ValueError, IndexError]

    class StackEntry(NamedTuple):
        function_name: Optional[str]
        line_number: int

    def __init__(self, message: Optional[str]):
        super().__init__(message)
        self.stack_entries: List[RuntimeError.StackEntry] = []
        self.next_name: Optional[str] = None

    def add_stack_entry(self, func_name: Optional[str], line_number: int):
        entry = RuntimeError.StackEntry(func_name, line_number)
        self.stack_entries.append(entry)

    @staticmethod
    def get_exception_line(e: Exception) -> int:
        for frame in reversed(traceback.extract_tb(e.__traceback__)):
            if frame.filename == "<string>":
                return frame.lineno - 1

    @staticmethod
    def update_error(func_name: Optional[str], e: Exception) -> "RuntimeError":
        if isinstance(e, RuntimeError):
            exp = RuntimeError(str(e))
            exp.stack_entries = e.stack_entries[::]
            exp.next_name = e.next_name
        else:
            if not any(isinstance(e, Exp) for Exp in RuntimeError.ACCEPTED_EXCEPTIONS):
                raise e
            exp = RuntimeError(str(e))
        line_num = RuntimeError.get_exception_line(e)
        if line_num is not None:
            exp.add_stack_entry(exp.next_name if func_name is None else func_name,
                                line_num)
        else:
            exp.next_name = func_name
        return exp



from typing import Tuple, List

from cubelang.cube import ICubeSide
from cubelang.orientation import Color
from cubelang.sides import Component
from cubelang.pattern import Pattern


class MockCubeSide(ICubeSide[None]):
    def __init__(self, colors: List[List[Color]]):
        super().__init__()
        self._colors = colors

    @property
    def rows(self) -> int:
        return len(self._colors)

    @property
    def columns(self) -> int:
        return len(self._colors[0])

    def __getitem__(self, item: Tuple[int, int]) -> Component[None]:
        i, j = item
        return Component(self._colors[i][j], None)

    def __setitem__(self, key: Tuple[int, int], value: Component[None]) -> None:
        raise NotImplementedError()


COLORS = [
    [Color.RED, Color.GREEN, Color.BLUE],
    [Color.WHITE, Color.WHITE, Color.ORANGE],
    [Color.ORANGE, Color.YELLOW, Color.YELLOW]
]


def test_pattern_match_correct():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, Color.GREEN, None],
        ["g1", "g1", None],
        [None, "g2", "g2"]
    ])
    match = pattern.match(side, {"g1": Color.WHITE})
    assert match == {"g2": Color.YELLOW}


def test_pattern_match_wrong_color():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, Color.RED, None],
        ["g1", "g1", None],
        [None, None, None]
    ])
    assert pattern.match(side, dict()) is None


def test_pattern_match_wrong_known():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, Color.GREEN, None],
        ["g1", Color.WHITE, None],
        [None, "g2", "g2"]
    ])
    assert pattern.match(side, {"g1": Color.YELLOW}) is None


def test_pattern_match_wrong_found():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, None, None],
        ["g1", None, None],
        ["g1", None, None]
    ])
    assert pattern.match(side, dict()) is None


def test_pattern_multiple_fount():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, None, None],
        ["g1", "g2", None],
        [None, None, None]
    ])
    assert pattern.match(side, dict()) is None


def test_pattern_multiple():
    side = MockCubeSide(COLORS)
    pattern = Pattern([
        [None, None, None],
        ["g1", "g2", None],
        [None, None, None]
    ])
    assert pattern.match(side, {"g1": Color.WHITE}) is None



from typing import Iterator, Dict
from string import whitespace

from .actions import Turn, Action, Rotate
from .orientation import Side


SIDE_LETTERS: Dict[str, Side] = {
    "L": Side.LEFT,
    "R": Side.RIGHT,
    "F": Side.FRONT,
    "B": Side.BACK,
    "U": Side.TOP,
    "D": Side.BOTTOM
}

ROTATE_LETTERS: Dict[str, Side] = {
    "X": Side.RIGHT,
    "Y": Side.TOP,
    "Z": Side.FRONT
}

ROTATE_REVERSE: Dict[Side, str] = {
    Side.RIGHT: "X", Side.LEFT: "X",
    Side.TOP: "Y", Side.BOTTOM: "Y",
    Side.FRONT: "Z", Side.BACK: "Z"
}


class ParsingError(Exception):
    def __init__(self, message: str, column: int) -> None:
        super(ParsingError, self).__init__(message)
        self.column = column


class ParsingStateMachine:
    def __init__(self):
        self.actions = []
        self.column = 0

        self.action_type = None
        self.amount = 1

        self.current_number = 0
        self.number_present = False
        self.numbers = []

    def yield_action(self):
        if self.action_type is None:
            return
        if self.action_type in ROTATE_LETTERS:
            side = ROTATE_LETTERS[self.action_type]
            if self.amount == 3:
                side = side.opposite()
            self.actions.append(Rotate(side, self.amount == 2))
        else:
            action = Turn(SIDE_LETTERS[self.action_type],
                          self.numbers if len(self.numbers) > 0 else 1,
                          self.amount)
            self.actions.append(action)

        self.action_type = None
        self.numbers = []
        self.amount = 1
        self.number_present = False

    def _unexpected(self, char: str):
        if char == "\n":
            raise ParsingError(f"Unexpected end at {self.column + 1}", self.column)
        else:
            raise ParsingError(f"Unexpected character: '{char}' at {self.column + 1}", self.column)

    def state_action_type(self, char: str):
        self.yield_action()
        if char == "\n":
            return True, None
        elif char not in SIDE_LETTERS and char not in ROTATE_LETTERS:
            self._unexpected(char)
        else:
            self.action_type = char
            return True, self.state_action_spec

    def state_action_spec(self, char: str):
        if char in SIDE_LETTERS or char in ROTATE_LETTERS:
            return False, self.state_action_type
        if char == "'":
            self.amount = 3
            return True, self.state_range_start
        elif char == "2":
            self.amount = 2
            return True, self.state_range_start
        elif char == "[":
            return False, self.state_range_start
        elif char == "\n":
            return True, None
        else:
            self._unexpected(char)

    def state_range_start(self, char: str):
        if char == "[":
            if self.action_type in ROTATE_LETTERS:
                self._unexpected(char)
            return True, self.state_range_number
        else:
            return False, self.state_action_type

    def state_range_number(self, char: str):
        def next_number(force_number: bool = False):
            if not self.number_present and force_number:
                self._unexpected(char)
            elif self.number_present:
                self.numbers.append(self.current_number)
            self.current_number = 0
            self.number_present = False

        if char.isdigit():
            self.current_number = self.current_number * 10 + int(char)
            self.number_present = True
        elif char == ":":
            if not self.number_present and len(self.numbers) > 0:
                self._unexpected(char)
            next_number()
            self.numbers.append(...)
        elif char == ",":
            next_number(True)
        elif char == "]":
            if not(self.number_present or (len(self.numbers) > 0 and self.numbers[-1] == Ellipsis)):
                self._unexpected(char)
            next_number()
            return True, self.state_action_type
        else:
            self._unexpected(char)
        return True, self.state_range_number

    def parse(self, algorithm: str):
        algorithm = algorithm.translate({ord(x): None for x in whitespace}) + "\n"
        state = self.state_action_type
        self.column = 0

        while self.column < len(algorithm):
            goto_next, state = state(algorithm[self.column])
            if goto_next:
                self.column += 1
        self.yield_action()


def parse_actions(algorithm: str) -> Iterator[Action]:
    sm = ParsingStateMachine()
    sm.parse(algorithm)
    return sm.actions



from typing import Dict, Any, List, Callable

from ..compiler.types import Type, Function, Bool
from ..compiler.stack import Stack
from ..execution.rt_function import RuntimeFunction


class Library:
    def __init__(self):
        self.global_values: Dict[str, Type] = dict()
        self.exec_globals: Dict[str, Any] = dict()

    def add_function(self, name: str, function: Callable, arguments: List[Type], return_type: Type):
        if name in self.global_values:
            function = self.global_values[name]
            assert isinstance(function, Function)
            function.prepend_overload(arguments, return_type)
        else:
            self.global_values[name] = Function((arguments, return_type))
            self.exec_globals[name] = function

    def function(self, arguments: List[Type], return_type: Type):
        def wrapper(function):
            if isinstance(function, RuntimeFunction):
                name = function.name
            else:
                name = function.__name__
                function = RuntimeFunction(name, function)
            self.add_function(name, function, arguments, return_type)
            return function
        return wrapper

    def add_value(self, name: str, value_type: Type, value: Any):
        self.global_values[name] = value_type
        self.exec_globals[name] = value

    def initialize_stack(self, stack: Stack):
        for name, value_type in self.global_values.items():
            stack.add_global(name, value_type)


stdlib = Library()
stdlib.add_value("true", Bool, True)
stdlib.add_value("false", Bool, False)


from .collections import *
from .math import *



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Callable, Optional, Union, Any

TIn = TypeVar("TIn")
TOut = TypeVar("TOut")


class PostprocessorBase(ABC, Generic[TIn, TOut]):
    def __init__(self):
        self.callback: Optional[Callable[[TOut], None]] = None
        self.done_callback: Optional[Callable[[], None]] = None

    @abstractmethod
    def process(self, value: TIn):
        pass

    def done(self):
        if self.done_callback is not None:
            self.done_callback()

    def _return(self, value: TOut):
        if self.callback is not None:
            self.callback(value)


def chain(*stages: Union[PostprocessorBase, Callable[[Any], None]]):
    if len(stages) < 2:
        return
    for i in range(1, len(stages)):
        if isinstance(stages[i], PostprocessorBase):
            stages[i - 1].callback = stages[i].process
            stages[i - 1].done_callback = stages[i].done
        else:
            stages[i - 1].callback = stages[i]
            if i != len(stages) - 1:
                raise ValueError("A callable can only be supplied into the "
                                 "chain function if it is the last argument")



from typing import List

from cubelang.cube import Cube, shift_list
from cubelang.sides import CubeSide, CubeSideView, ICubeSide
from cubelang.orientation import Orientation, Side, Color
from cubelang.parser import parse_actions
from cubelang.pattern import Pattern

from unittest.mock import MagicMock
import pytest


def test_shift_list() -> None:
    lst = [1, 2, 3, 4, 5]
    assert shift_list(lst) == [2, 3, 4, 5, 1]


def test_shift_list_2() -> None:
    lst = [1, 2, 3, 4, 5]
    assert shift_list(lst, 2) == [3, 4, 5, 1, 2]


def test_get_side():
    orientation: Orientation = Orientation(Side.LEFT, Side.TOP)
    orientation.get_side_rotation = MagicMock()
    orientation.get_side_rotation.return_value = 0

    cube = Cube((3, 3, 3))
    side = cube.get_side(orientation)
    assert isinstance(side, CubeSide)
    assert side == cube.sides[Side.LEFT]


def test_get_side_rotated():
    orientation: Orientation = Orientation(Side.LEFT, Side.TOP)
    orientation.get_side_rotation = MagicMock()
    orientation.get_side_rotation.return_value = 3

    cube = Cube((3, 3, 3))
    side = cube.get_side(orientation)
    assert isinstance(side, CubeSideView)
    assert side.side == cube.sides[Side.LEFT]
    assert side.rotation == 3


def _set_side_colors(side: ICubeSide, colors: List[List[Color]]) -> None:
    for i, row in enumerate(colors):
        for j, color in enumerate(row):
            side.colors[i, j] = color


def side_to_string(side: ICubeSide):
    out = []
    for i in range(side.rows):
        for j in range(side.columns):
            color = side.colors[i, j].name[0].upper()
            out.append(color)
        out.append("/")
    return "".join(out[:-1])


def data_to_string(side: ICubeSide):
    out = []
    for i in range(side.rows):
        out.append(" ".join(str(side[i, j].data) for j in range(side.columns)))
        out.append("/")
    return "".join(out[:-1])


def assert_cube(cube: Cube, front: str, right: str, back: str, left: str,
                top: str, bottom: str) -> None:
    orientation = Orientation()
    assert front == side_to_string(cube.get_side(orientation))
    assert right == side_to_string(cube.get_side(orientation.to_right))
    assert back == side_to_string(cube.get_side(orientation.to_right.to_right))
    assert left == side_to_string(cube.get_side(orientation.to_left))
    assert top == side_to_string(cube.get_side(orientation.to_top))
    assert bottom == side_to_string(cube.get_side(orientation.to_bottom))


@pytest.fixture()
def sample_cube() -> Cube:
    cube = Cube((3, 3, 3))
    orientation = Orientation()
    assert orientation.get_side_rotation() == 0

    red = cube.get_side(orientation)
    _set_side_colors(red, [
        [Color.ORANGE, Color.RED, Color.GREEN],
        [Color.YELLOW, Color.RED, Color.GREEN],
        [Color.BLUE, Color.WHITE, Color.GREEN]
    ])

    green = cube.get_side(orientation.to_right)
    _set_side_colors(green, [
        [Color.RED, Color.ORANGE, Color.ORANGE],
        [Color.RED, Color.GREEN, Color.WHITE],
        [Color.RED, Color.GREEN, Color.WHITE]
    ])

    orange = cube.get_side(orientation.to_left.to_left)
    _set_side_colors(orange, [
        [Color.YELLOW, Color.YELLOW, Color.RED],
        [Color.ORANGE, Color.ORANGE, Color.WHITE],
        [Color.ORANGE, Color.ORANGE, Color.GREEN]
    ])

    blue = cube.get_side(orientation.to_left)
    _set_side_colors(blue, [
        [Color.YELLOW, Color.YELLOW, Color.YELLOW],
        [Color.BLUE, Color.BLUE, Color.RED],
        [Color.ORANGE, Color.WHITE, Color.WHITE]
    ])

    yellow = cube.get_side(orientation.to_top)
    _set_side_colors(yellow, [
        [Color.BLUE, Color.GREEN, Color.GREEN],
        [Color.BLUE, Color.YELLOW, Color.YELLOW],
        [Color.BLUE, Color.BLUE, Color.WHITE]
    ])

    white = cube.get_side(orientation.to_bottom)
    _set_side_colors(white, [
        [Color.RED, Color.GREEN, Color.YELLOW],
        [Color.RED, Color.WHITE, Color.ORANGE],
        [Color.WHITE, Color.BLUE, Color.BLUE]
    ])

    assert_cube(cube, "ORG/YRG/BWG", "ROO/RGW/RGW", "YYR/OOW/OOG", "YYY/BBR/OWW",
                "BGG/BYY/BBW", "RGY/RWO/WBB")
    return cube


def test_rotation_rightmost(sample_cube: Cube) -> None:
    orientation = Orientation(Side.FRONT, Side.RIGHT)
    sample_cube.turn_vertical(orientation, -1, 1)

    assert_cube(sample_cube, "ORG/YRG/OWW", "ROO/RGW/BWG", "YYR/OOW/RGW", "YYY/BBR/OOG",
                "BGG/BYY/BBW", "WRR/BWG/BOY")


def test_rotation_leftmost(sample_cube: Cube) -> None:
    orientation = Orientation(Side.BACK, Side.BOTTOM)
    sample_cube.turn_vertical(orientation, 1, 3)

    assert_cube(sample_cube, "BRG/BRG/BWG", "ROO/RGW/RGW", "YYW/OOR/OOR", "OBY/WBY/WRY",
                "GGG/WYY/RBW", "OGY/YWO/BBB")


def test_rotation_horizontal(sample_cube: Cube) -> None:
    orientation = Orientation(Side.RIGHT, Side.FRONT)
    sample_cube.turn_horizontal(orientation, 1, 1)

    assert_cube(sample_cube, "GGG/RRW/OYB", "YOO/GGW/RGW", "YYR/OOW/OOG", "YYW/BBB/OWB",
                "BGG/BYY/RRR", "YRW/RWO/WBB")


def test_rotation_horizontal_last(sample_cube: Cube) -> None:
    orientation = Orientation(Side.BOTTOM, Side.LEFT)
    sample_cube.turn_horizontal(orientation, -1, 3)

    assert_cube(sample_cube, "ORG/YRY/BWW", "OWW/OGG/RRR", "BYR/OOW/YOG", "YYY/BBR/OWW",
                "BGO/BYO/BBY", "RGG/RWG/WBG")


def test_rotation_slice(sample_cube: Cube) -> None:
    orientation = Orientation(Side.RIGHT, Side.BACK)
    sample_cube.turn_slice(orientation, 1, 1)

    assert_cube(sample_cube, "ORY/YRO/BWB", "RRR/GGO/WWO", "WYR/YOW/GOG", "YYY/BBR/OWW",
                "BGG/BYG/BBG", "RGO/RWO/WBY")


def test_data_set(sample_cube: Cube) -> None:
    orientation = Orientation(Side.RIGHT, Side.BACK)

    sample_cube.set_data(orientation, 1, 1, "a")
    sample_cube.set_data(orientation, 0, 0, "b")
    sample_cube.set_data(orientation, 1, 2, "c")
    sample_cube.set_data(orientation, 2, 1, "d")

    def get_side(orient: Orientation) -> str:
        return data_to_string(sample_cube.get_side(orient))

    orientation = Orientation(Side.FRONT, Side.TOP)
    assert get_side(orientation) == "None None None/None None d/None None None"
    assert get_side(orientation.to_top) == "None None b/None None None/None None None"
    assert get_side(orientation.to_right) == "None None b/d a None/None c None"
    assert get_side(orientation.to_right.to_right) == "b None None/None None None/None None None"
    assert get_side(orientation.to_bottom) == "None None None/None None c/None None None"


def test_data_rotation(sample_cube: Cube) -> None:
    orientation = Orientation(Side.FRONT, Side.TOP)
    sample_cube.set_data(orientation, 0, 2, "b")
    sample_cube.set_data(orientation, 1, 2, "a")
    sample_cube.turn_vertical(orientation, 3, 1)

    def get_side(orient: Orientation) -> str:
        return data_to_string(sample_cube.get_side(orient))

    assert get_side(orientation) == "None None None/None None None/None None None"
    assert get_side(orientation.to_right) == "None a b/None None None/None None None"
    assert get_side(orientation.to_top) == "None None b/None None a/None None None"


def test_iterate() -> None:
    def orient_to_str(side: Side, i: int, j: int) -> str:
        return f"{side.name[0].upper()}{i}:{j}"

    cube = Cube((3, 4, 4))
    result = set(map(lambda x: orient_to_str(*x), cube.iterate_components()))
    assert result == {"F0:0", "F0:1", "F0:2", "F1:0", "F1:1", "F1:2", "F2:0", "F2:1", "F2:2", "F3:0", "F3:1", "F3:2",
                      "R0:1", "R0:2", "R1:1", "R1:2", "R2:1", "R2:2", "R3:1", "R3:2",
                      "B0:0", "B0:1", "B0:2", "B1:0", "B1:1", "B1:2", "B2:0", "B2:1", "B2:2", "B3:0", "B3:1", "B3:2",
                      "L0:1", "L0:2", "L1:1", "L1:2", "L2:1", "L2:2", "L3:1", "L3:2",
                      "T1:1", "T2:1", "B1:1", "B2:1"}


@pytest.mark.parametrize("side, i, j, a, b, c", [
    (Side.FRONT, 0, 0, 0, 0, 0),
    (Side.FRONT, 1, 3, 3, 1, 0),
    (Side.LEFT, 2, 3, 0, 2, 1),
    (Side.RIGHT, 0, 1, 3, 0, 1),
    (Side.TOP, 3, 2, 2, 0, 1),
    (Side.BOTTOM, 3, 1, 1, 2, 3),
    (Side.BACK, 1, 1, 2, 1, 4)
])
def test_get_absolute(side, i, j, a, b, c) -> None:
    cube = Cube((4, 5, 3))
    x, y, z = cube.get_absolute_coordinates(side, i, j)
    assert x == a
    assert y == b
    assert z == c


@pytest.mark.parametrize("side, i, j, a, b, c", [
    (Side.FRONT, 0, 0, 0, 0, 0), (Side.FRONT, 1, 0, 0, 1, 0), (Side.FRONT, 0, 1, 1, 0, 0), (Side.FRONT, 1, 1, 1, 1, 0),
    (Side.RIGHT, 0, 0, 1, 0, 0), (Side.RIGHT, 1, 0, 1, 1, 0), (Side.RIGHT, 0, 1, 1, 0, 1), (Side.RIGHT, 1, 1, 1, 1, 1),
    (Side.LEFT,  0, 0, 0, 0, 1), (Side.LEFT,  1, 0, 0, 1, 1), (Side.LEFT,  0, 1, 0, 0, 0), (Side.LEFT,  1, 1, 0, 1, 0),
    (Side.TOP,   0, 0, 0, 0, 1), (Side.TOP,   1, 0, 0, 0, 0), (Side.TOP,   0, 1, 1, 0, 1), (Side.TOP,   1, 1, 1, 0, 0),
    (Side.BACK,  0, 0, 1, 0, 1), (Side.BACK,  1, 0, 1, 1, 1), (Side.BACK,  0, 1, 0, 0, 1), (Side.BACK,   1, 1, 0, 1, 1),
])
def test_get_absolute_2(side, i, j, a, b, c) -> None:
    cube = Cube((2, 2, 2))
    x, y, z = cube.get_absolute_coordinates(side, i, j)
    assert x == a
    assert y == b
    assert z == c


def test_cube_flip_flop():
    cube = Cube((3, 3, 3))
    assert_cube(cube, "RRR/RRR/RRR", "GGG/GGG/GGG", "OOO/OOO/OOO", "BBB/BBB/BBB", "YYY/YYY/YYY", "WWW/WWW/WWW")

    for action in parse_actions("RUR'U'"):
        action.perform(cube, Orientation())
    assert_cube(cube, "RRW/RRY/RRR", "GGY/OGG/YGG", "OGG/OOO/OOO", "OBB/BBB/BBB", "YYB/YYR/YYR", "WWG/WWW/WWW")

    for _ in range(5):
        for action in parse_actions("RUR'U'"):
            action.perform(cube, Orientation())
    assert_cube(cube, "RRR/RRR/RRR", "GGG/GGG/GGG", "OOO/OOO/OOO", "BBB/BBB/BBB", "YYY/YYY/YYY", "WWW/WWW/WWW")


def test_orient_full():
    cube = Cube((3, 3, 3))
    for action in parse_actions("RUR'U'"):
        action.perform(cube, Orientation())

    match = cube.orient(Orientation(),
                        top=Pattern([[Color.WHITE, None, None], [None, "a", None], ["a", None, None]]),
                        front=Pattern([[None, None, None], [None, None, Color.ORANGE], [None, None, None]]),
                        right=Pattern([[None, None, None], [Color.YELLOW, None, None], ["a", None, None]]),
                        back=Pattern([[None, None, None], [None, None, None], [Color.BLUE, None, None]]),
                        left=Pattern([[None, Color.ORANGE, None], [None, None, None], [None, Color.GREEN, None]]),
                        bottom=Pattern([[Color.BLUE, None, Color.RED], [None, None, None], [None, None, None]]))
    assert Orientation(Side.RIGHT, Side.BOTTOM) == match


def test_orient_keep():
    cube = Cube((3, 3, 3))
    for action in parse_actions("RUR'U'"):
        action.perform(cube, Orientation())

    match = cube.orient(Orientation(), keeping=Side.LEFT,
                        bottom=Pattern([[Color.WHITE, None, None], [None, None, None], [None, None, None]]))
    assert Orientation(Side.FRONT, Side.TOP) == match


def test_orient_not_matching_side():
    cube = Cube((3, 3, 3))
    for action in parse_actions("RUR'U'"):
        action.perform(cube, Orientation())
    match = cube.orient(Orientation(),
                        front=Pattern([[Color.WHITE, Color.GREEN, None], [None, None, None], [None, None, None]]))
    assert match is None


def test_orient_not_matching_groups():
    cube = Cube((3, 3, 3))
    for action in parse_actions("RUR'U'"):
        action.perform(cube, Orientation())

    match = cube.orient(Orientation(),
                        front=Pattern([["a", "a", "b"], [None, None, None], [None, None, None]]),
                        right=Pattern([["a", None, None], [None, None, None], ["a", None, "b"]]))
    assert match is None



from . import stdlib
from ..compiler.types import Real, Integer, List, Set, Bool
import math


stdlib.add_function("round", round, [Real], Integer)
stdlib.add_function("floor", math.floor, [Real], Integer)
stdlib.add_function("ceil", math.ceil, [Real], Integer)

stdlib.add_function("sqrt", math.sqrt, [Real], Real)
stdlib.add_function("pow", math.pow, [Real, Real], Real)
stdlib.add_function("pow", math.pow, [Integer, Integer], Integer)

stdlib.add_function("max", max, [Real, Real], Real)
stdlib.add_function("max", max, [Integer, Integer], Integer)
stdlib.add_function("max", max, [List(Real)], Real)
stdlib.add_function("max", max, [List(Integer)], Integer)
stdlib.add_function("max", max, [Set(Real)], Real)
stdlib.add_function("max", max, [Set(Integer)], Integer)

stdlib.add_function("min", min, [Real, Real], Real)
stdlib.add_function("min", min, [Integer, Integer], Integer)
stdlib.add_function("min", min, [List(Real)], Real)
stdlib.add_function("min", min, [List(Integer)], Integer)
stdlib.add_function("min", min, [Set(Real)], Real)
stdlib.add_function("min", min, [Set(Integer)], Integer)


@stdlib.function([Real], Integer)
def sign(x: float) -> int:
    if x > 1e-7:
        return 1
    elif x < -1e-7:
        return -1
    else:
        return 0

stdlib.add_function("not", lambda x: not x, [Bool], Bool)



from collections import namedtuple

from cubelang.cube import Cube
from cubelang.orientation import Side, Orientation
from cubelang.actions import Rotate, Turn, TurningType

import pytest
from typing import List
from unittest.mock import MagicMock


@pytest.mark.parametrize("around, front, top, twice", [
    (Side.FRONT, Side.FRONT, Side.LEFT, False),
    (Side.BACK, Side.FRONT, Side.RIGHT, False),
    (Side.RIGHT, Side.BOTTOM, Side.FRONT, False),
    (Side.LEFT, Side.TOP, Side.BACK, False),
    (Side.TOP, Side.RIGHT, Side.TOP, False),
    (Side.BOTTOM, Side.LEFT, Side.TOP, False),

    (Side.FRONT, Side.FRONT, Side.BOTTOM, True),
    (Side.LEFT, Side.BACK, Side.BOTTOM, True)
])
def test_rotations(around: Side, front: Side, top: Side, twice: bool) -> None:
    orientation = Orientation(Side.FRONT, Side.TOP)
    action = Rotate(around, twice)
    # noinspection PyTypeChecker
    orientation = action.perform(None, orientation)

    assert orientation.front == front
    assert orientation.top == top


# noinspection PyTypeChecker
@pytest.mark.parametrize("around", [x for x in Side])
def test_rotation_2(around: Side):
    action1 = Rotate(around, False)
    action2 = Rotate(around.opposite(), False)

    initial = Orientation(Side.LEFT, Side.BACK)
    orientation = action1.perform(None, initial)
    orientation = action2.perform(None, orientation)

    assert orientation == initial


def test_from_turn_steps():
    turns = [Side.RIGHT, Side.FRONT, Side.FRONT,
             Side.TOP, Side.TOP, Side.TOP,
             Side.RIGHT, Side.RIGHT, Side.RIGHT, Side.RIGHT, Side.TOP]
    actions = list(Rotate.from_turn_steps(turns))
    expected_sides = [Side.RIGHT, Side.FRONT, Side.BOTTOM, Side.TOP]
    expected_twice = [False, True, False, False, False]
    assert len(actions) == len(expected_sides)
    for action, side, twice in zip(actions, expected_sides, expected_twice):
        assert action.axis_side == side
        assert action.twice == twice


class CubeMock(object):
    SideMock = namedtuple("SideMock", ["columns", "rows"])

    def __init__(self):
        self.turn_slice = None
        self.turn_vertical = None
        self.turn_horizontal = None

    def get_side(self, _orientation):
        return CubeMock.SideMock(3, 3)


@pytest.mark.parametrize("side, func, out_sides, out_amount", [
    (Side.FRONT, "turn_slice", [1, 2], 1),
    (Side.BACK, "turn_slice", [2, 3], 3),
    (Side.RIGHT, "turn_vertical", [2, 3], 1),
    (Side.LEFT, "turn_vertical", [1, 2], 3),
    (Side.TOP, "turn_horizontal", [1, 2], 3),
    (Side.BOTTOM, "turn_horizontal", [2, 3], 1)
])
def test_turning_vertical(side: Side, func: str, out_sides: List[int], out_amount: int) -> None:
    cube: Cube = CubeMock()
    mock = MagicMock()
    setattr(cube, func, mock)

    orientation = Orientation()
    action = Turn(side, [1, 2], 1)
    assert action.perform(cube, orientation) == orientation
    assert len(mock.call_args_list) == len(out_sides)
    print(mock.call_args_list[0])
    for args, side in zip(mock.call_args_list, out_sides):
        arg_orientation, arg_index, arg_turn = tuple(args)[0]
        assert arg_orientation == orientation
        assert arg_index == side
        assert arg_turn == out_amount


@pytest.mark.parametrize("side, turn, res_side, res_sides, res_turns", [
    (Side.TOP,   Side.FRONT, TurningType.VERTICAL, -1, 1),
    (Side.RIGHT, Side.FRONT, TurningType.HORIZONTAL, -1, 1),

    (Side.RIGHT, Side.TOP, TurningType.SLICE, 1, 1),
    (Side.FRONT, Side.TOP, TurningType.VERTICAL, 1, 3),

    (Side.FRONT, Side.RIGHT, TurningType.HORIZONTAL, 1, 3),
    (Side.TOP, Side.RIGHT, TurningType.SLICE, -1, 3),
])
def test_turning_transform(side: Side, turn: Side, res_side: TurningType,
                           res_sides: int, res_turns: int):
    action = Turn(side, [1], 1)
    transformed = action._transform(turn)
    assert transformed.type == res_side
    assert transformed.turns == res_turns
    assert transformed.indices[0] == res_sides


@pytest.mark.parametrize("orientation, action, type, indices", [
    (Orientation(Side.BACK, Side.LEFT), Turn(Side.RIGHT, 1, 3), TurningType.HORIZONTAL, -1),
    (Orientation(Side.TOP, Side.BACK), Turn(Side.TOP, 1, 1), TurningType.SLICE, -1),
    (Orientation(Side.RIGHT, Side.BOTTOM), Turn(Side.FRONT, 1, 1), TurningType.VERTICAL, -1)
])
def test_orientation_changes(orientation: Side, action: Turn, type: TurningType,
                             indices: int):
    new_action = action.from_orientation(orientation)
    assert new_action.type == type
    assert new_action.indices[0] == indices


@pytest.mark.parametrize("indices, expected", [
    ([2], {2}),
    ([2, 4], {2, 4}),
    ([2, ..., 4], {2, 3, 4}),
    ([1, ..., 3, 5], {1, 2, 3, 5}),
    ([1, ..., 3,  2, ..., 5], {1, 2, 3, 4, 5}),
    ([..., 3, 5], {1, 2, 3, 5}),
    ([1, 3, ...], {1, 3, 4, 5}),
    ([-2, ..., -1], {4, 5}),
    ([..., -3], {1, 2, 3}),
    ([-3, ...], {3, 4, 5})
])
def test_normalize_indices(indices, expected):
    assert expected == Turn.normalize_indices(indices, 5)


def _iterate_orientations():
    for side in Side:
        orientation = Orientation.regular(side)
        for i in range(4):
            yield orientation
            orientation = orientation.rotate_clockwise()


def cartesian(a, b):
    for x in a:
        for y in b:
            yield x, y


@pytest.mark.parametrize("orientation, turn", cartesian(
    list(_iterate_orientations()), [
        Turn(Side.LEFT, 1), Turn(Side.LEFT, 1, 3),
        Turn(Side.RIGHT, 1), Turn(Side.RIGHT, 1, 3),
        Turn(Side.TOP, 1), Turn(Side.TOP, 1, 3),
        Turn(Side.BOTTOM, 1), Turn(Side.BOTTOM, 1, 3),
        Turn(Side.FRONT, 1), Turn(Side.FRONT, 1, 3),
        Turn(Side.BACK, 1), Turn(Side.BACK, 1, 3)]))
def test_orientation_transform(orientation, turn):
    cube1 = Cube((3, 3, 3))
    cube2 = Cube((3, 3, 3))

    turn.perform(cube1, orientation)
    origin = Orientation()
    turn.from_orientation(orientation, origin).perform(cube2, origin)

    for side in Side:
        o = Orientation.regular(side)
        assert str(cube1.get_side(o).colors) == str(cube2.get_side(o).colors)



from . import stdlib
from ..compiler.types import Integer, List, Set, T, Void, Bool
import typing

CollectionType = typing.Union[typing.Set, typing.List]


# Set functions

@stdlib.function([Set(T)], Integer)
@stdlib.function([List(T)], Integer)
def size(collection: CollectionType):
    """ Returns a number of elements in the collection. """
    return len(collection)


@stdlib.function([Set(T), T], Void)
def add(collection: typing.Set, value) -> None:
    """ Adds a new value to the set. """
    collection.add(value)


@stdlib.function([Set(T), T], Bool)
def remove(collection: typing.Set, value) -> bool:
    """ Removes a value from the set. Returns true if the value were present,
    false otherwise. """
    try:
        collection.remove(value)
        return True
    except KeyError:
        return False


@stdlib.function([Set(T), T], Bool)
@stdlib.function([List(T), T], Bool)
def contains(collection: CollectionType, value) -> bool:
    """ Returns true if the value is present in the collection, false otherwise. """
    return value in collection


@stdlib.function([Set(T), T], Void)
@stdlib.function([List(T), T], Void)
def clear(collection: CollectionType) -> None:
    """ Removes all values from the collection. """
    collection.clear()


@stdlib.function([T, ...], Set(T))
def set_of(*args):
    return set(args)


# List functions

@stdlib.function([List(T), T], Void)
def add_first(collection: typing.List, value) -> None:
    """ Adds an elements to the begining of a list. Its index will be 0. """
    collection.insert(0, value)


@stdlib.function([List(T), T], Void)
def add_last(collection: typing.List, value) -> None:
    """ Adds an element to the end of a list. Its index will be one less
    than the size of a list. """
    collection.append(value)


@stdlib.function([List(T), Integer, T], Void)
def add_at(collection: typing.List, index: int, value) -> None:
    """ Adds an elements to the list at a specified location. """
    if index < 0:
        raise ValueError("List index must be zero or a positive integer")
    collection.insert(index, value)


@stdlib.function([List(T)], T)
def remove_first(collection: typing.List):
    """ Removes and returns the first element of a list. """
    if len(collection) == 0:
        raise ValueError("The list is empty. There is notthing to remove")
    return collection.pop(0)


@stdlib.function([List(T)], T)
def remove_last(collection: typing.List):
    """ Removes and returns the last element of a list. """
    if len(collection) == 0:
        raise ValueError("The list is empty. There is notthing to remove")
    return collection.pop()


@stdlib.function([List(T), Integer], T)
def remove_at(collection: typing.List, index: int):
    """ Removes and returns an element of a list at specified location. """
    if index < 0 or index >= len(collection):
        raise ValueError("List index is out of range")
    element = collection[index]
    del collection[index]
    return element


@stdlib.function([Integer, T], List(T))
def new_list(size: int, value):
    """ Creates a new list of length `size` and fills it with `value`s. """
    if size < 0:
        raise ValueError("List size must be a positive integer")
    return [value] * size


@stdlib.function([List(T), T], Integer)
def index_of(array: list, value) -> int:
    """ Returns index of an element in a list if it is present, or -1 otherwise. """
    try:
        return array.index(value)
    except ValueError:
        return -1


@stdlib.function([T, ...], List(T))
def list_of(*args):
    return list(args)



from typing import List, Tuple, TypeVar, Dict

from ..actions import Action
from ..parser import ParsingError, parse_actions
from ..orientation import Color

from argparse import ArgumentTypeError

T = TypeVar("T")


def truncate_string_around(value: str, column: int, left_offset: int = 20, right_offset: int = 20,
                           truncation: str = "...") -> Tuple[str, int]:
    assert len(truncation) < left_offset
    assert len(truncation) < right_offset

    if column > left_offset:
        value = truncation + value[column - left_offset + len(truncation):]
        column = left_offset

    if len(value) - column - 1 > right_offset:
        value = value[:column + right_offset - len(truncation) + 1] + truncation

    return value, column


def _get_syntax_error_message(message: str, text: str, position: int):
    text = text.replace("\n", " ").replace("\t", " ")
    value, column = truncate_string_around(text, position)
    return "".join([message, "\n", " " * 7, value, "\n", " " * (7 + column), "^"])


def integer_type(min_value: int):
    def type(value: str):
        try:
            val = int(value)
            if val < min_value:
                raise ArgumentTypeError(f"the minimum value is {min_value}")
            return val
        except ValueError:
            raise ArgumentTypeError(f"invalid integer value: '{value}'")
    return type


def dict_type(dictionary: Dict[str, T]):
    def type(value: str) -> T:
        if value in dictionary:
            return dictionary[value]
        else:
            keys = list(dictionary.keys())
            values = ", ".join(f"'{x}'" for x in keys[:-1]) + f" or '{keys[-1]}'"
            raise ArgumentTypeError(f"unknown value: '{value}'; expected either {values}")
    return type


def formula_type(value: str) -> List[Action]:
    try:
        return list(parse_actions(value))
    except ParsingError as e:
        raise ArgumentTypeError(_get_syntax_error_message(str(e), value, e.column))


SYMBOLS = {"R": Color.RED, "G": Color.GREEN, "B": Color.BLUE,
           "W": Color.WHITE, "Y": Color.YELLOW, "O": Color.ORANGE}


def side_colors_type(value: str):
    result: List[List[Color]] = []
    lines = value.upper().split("/")
    position = 0
    for line in lines:
        result_line: List[Color] = []
        if len(line) != len(lines[0]):
            raise ArgumentTypeError("inconsistent line length")
        for symbol in line:
            if symbol not in SYMBOLS:
                msg = _get_syntax_error_message(f"unknown color: `{symbol}`", value, position)
                raise ArgumentTypeError(msg)
            else:
                result_line.append(SYMBOLS[symbol])
            position += 1
        result.append(result_line)
        position += 1
    return result


def file_contents_type(value: str):
    try:
        with open(value) as file:
            return file.read()
    except IOError as e:
        raise ArgumentTypeError(f"cannot open a file: {e}")



from typing import Tuple
from cubelang.orientation import Side, Orientation, _normalize_turns

import pytest

T = Side.TOP
L = Side.LEFT
R = Side.RIGHT
D = Side.BOTTOM
F = Side.FRONT
B = Side.BACK


@pytest.mark.parametrize("current, left, top, right, bottom", [
    ((F, T), (L, T), (T, B), (R, T), (D, F)),
    ((F, R), (T, R), (R, B), (D, R), (L, F)),
    ((F, D), (R, D), (D, B), (L, D), (T, F)),
    ((F, L), (D, L), (L, B), (T, L), (R, F)),

    ((R, T), (F, T), (T, L), (B, T), (D, R)),
    ((R, B), (T, B), (B, L), (D, B), (F, R)),
    ((R, D), (B, D), (D, L), (F, D), (T, R)),
    ((R, F), (D, F), (F, L), (T, F), (B, R)),

    ((L, T), (B, T), (T, R), (F, T), (D, L)),
    ((L, B), (D, B), (B, R), (T, B), (F, L)),
    ((L, D), (F, D), (D, R), (B, D), (T, L)),
    ((L, F), (T, F), (F, R), (D, F), (B, L)),

    ((B, D), (L, D), (D, F), (R, D), (T, B)),
    ((B, L), (T, L), (L, F), (D, L), (R, B)),
    ((B, T), (R, T), (T, F), (L, T), (D, B)),
    ((B, R), (D, R), (R, F), (T, R), (L, B)),

    ((T, B), (L, B), (B, D), (R, B), (F, T)),
    ((T, L), (F, L), (L, D), (B, L), (R, T)),
    ((T, R), (B, R), (R, D), (F, R), (L, T)),
    ((T, F), (R, F), (F, D), (L, F), (B, T)),

    ((D, B), (R, B), (B, T), (L, B), (F, D)),
    ((D, L), (B, L), (L, T), (F, L), (R, D)),
    ((D, R), (F, R), (R, T), (B, R), (L, D)),
    ((D, F), (L, F), (F, T), (R, F), (B, D))
])
def test_rotations(current: Tuple[Side, Side], left: Tuple[Side, Side],
                   top: Tuple[Side, Side], right: Tuple[Side, Side],
                   bottom: Tuple[Side, Side]) -> None:
    current_orientation = Orientation(*current)

    assert current_orientation.to_top.front == top[0]
    assert current_orientation.to_top.top == top[1]

    assert current_orientation.to_right.front == right[0]
    assert current_orientation.to_right.top == right[1]

    assert current_orientation.to_bottom.front == bottom[0]
    assert current_orientation.to_bottom.top == bottom[1]

    assert current_orientation.to_left.front == left[0]
    assert current_orientation.to_left.top == left[1]


@pytest.mark.parametrize("front, top, rotation", [
    (F, T, 0), (F, R, 1), (F, D, 2), (F, L, 3),
    (R, T, 0), (R, B, 1), (R, D, 2), (R, F, 3),
    (B, T, 0), (B, L, 1), (B, D, 2), (B, R, 3),
    (L, T, 0), (L, F, 1), (L, D, 2), (L, B, 3),
    (T, B, 0), (T, R, 1), (T, F, 2), (T, L, 3),
    (D, L, 0), (D, F, 1), (D, R, 2), (D, B, 3)
])
def test_get_side_rotation(front: Side, top: Side, rotation: int):
    orientation = Orientation(front, top)
    assert orientation.get_side_rotation() == rotation


rotations = [
    (F, T, [R, D, L]),
    (R, T, [B, D, F]),
    (B, T, [L, D, R]),
    (L, T, [F, D, B]),
    (T, R, [F, L, B]),
    (D, L, [F, R, B])
]


@pytest.mark.parametrize("front, top, after_rotation", rotations)
def test_rotation_counterclockwise(front, top, after_rotation):
    orientation = Orientation(front, top).rotate_clockwise()
    for after_top in after_rotation[::-1]:
        assert orientation.top == after_top
        orientation = orientation.rotate_clockwise()
    assert orientation.top == top


@pytest.mark.parametrize("front, top, after_rotation", rotations)
def test_rotation_clockwise(front, top, after_rotation):
    orientation = Orientation(front, top).rotate_counterclockwise()
    for after_top in after_rotation:
        assert orientation.top == after_top
        orientation = orientation.rotate_counterclockwise()
    assert orientation.top == top


@pytest.mark.parametrize("front, top", [
    (Side.FRONT, Side.TOP),
    (Side.LEFT, Side.TOP),
    (Side.RIGHT, Side.TOP),
    (Side.BACK, Side.TOP),
    (Side.TOP, Side.BACK),
    (Side.BOTTOM, Side.FRONT)
])
def test_regular_orientation(front: Side, top: Side) -> None:
    orientation = Orientation.regular(front)
    assert orientation.front == front
    assert orientation.top == top


@pytest.mark.parametrize("keeping, expected", [
    (Side.FRONT, {(R, T), (R, F), (R, B), (R, D)}),
    (Side.BOTTOM, {(R, T), (F, T), (L, T), (B, T)}),
    (Side.LEFT, {(R, T), (T, L), (L, D), (D, R)}),
    (None, {(F, T), (F, R), (F, L), (F, D),  (R, T), (R, F), (R, B), (R, D),
            (B, T), (B, L), (B, R), (B, D),  (L, T), (L, F), (L, B), (L, D),
            (T, F), (T, L), (T, R), (T, B),  (D, F), (D, L), (D, R), (D, B)})
])
def test_iterate_orientations(keeping, expected):
    orientation = Orientation(Side.RIGHT, Side.TOP)
    rotations = list(orientation.iterate_rotations(keeping))
    assert len(expected) == len(rotations)
    assert {Orientation(a, b) for a, b in expected} == set(rotations)


@pytest.mark.parametrize("from_front, from_top, to_front, to_top, steps", [
    (F, T, F, T, []),
    (F, D, F, T, [F, F]),
    (R, B, F, T, [R, F  ]),
    (T, L, F, T, [T, T, T, F, F, F]),
    (D, R, F, T, [T, T, T, F]),
    (B, R, F, R, [T, T]),
    (R, B, F, D, [R, F, F, F]),
    (R, F, L, D, [T, T, F, F, F]),
    (F, T, L, D, [T, T, T, F, F]),
])
def test_turns_to_origin(from_front, from_top, to_front, to_top, steps):
    orientation = Orientation(from_front, from_top)
    actual = list(orientation.turns_to(Orientation(to_front, to_top)))
    assert actual == steps


def test_normalize_turns():
    stream = [R, R, R, R, L, L, T, D, D, D, L]
    expected = [R, R, T, T, R, R, R]
    assert expected == list(_normalize_turns(stream))



from collections.abc import Iterable
from pathlib import Path
from typing import Union, List, Callable, Dict, IO, Iterator, Tuple

from lark import Tree, Lark, Token

from .expression import Expression, ConditionExpression, WhileLoopExpression, \
    DoWhileLoopExpression, RepeatLoopExpression, ForLoopExpression, \
    CubeTurningExpression, CubeRotationExpression, FunctionDeclarationExpression
from .operators import BINARY_OPERATORS, BinaryOperator, operator_applicable
from .stack import Stack
from .types import Integer, Real, Type, Bool, Set, List as ListType, Void, \
    CollectionType, Function, Color, Side, Pattern
from .errors import assert_type, ValueTypeError, UnresolvedReferenceError, \
    FunctionArgumentsError, CompileTimeError

TYPE_NAMES = {"type_int": Integer, "type_real": Real, "type_bool": Bool,
              "type_color": Color, "type_side": Side, "type_pattern": Pattern}

Callback = Callable[[Tree, Stack], Union[Type, Expression]]

RESERVED_NAMES = {"int", "real", "boolean", "side", "color", "pattern", "list",
                  "set", "of", "func", "let", "return", "if", "then", "end",
                  "while", "do", "repeat", "for", "in", "orient"}


# TODO: Symbols aren't visible inside functions

class Parser:
    def __init__(self):
        path = Path(__file__).parents[1] / "data" / "syntax.lark"
        with open(str(path)) as f:
            grammar = Parser._generate_operators(BINARY_OPERATORS) + f.read()
        self.lark = Lark(grammar, parser="lalr", start="clause", propagate_positions=True)
        self.callbacks: Dict[str, Callback] = dict()

    @staticmethod
    def _generate_operators(operator_groups: List[List[BinaryOperator]]) -> str:
        lines = []
        for op_index, op in enumerate(operator_groups):
            this_name = "op_" + str(op_index)
            next_name = "op_" + str(op_index + 1) if op_index < len(operator_groups) - 1 else "op_if"
            lines.append(f"?{this_name}: {next_name} | ")
            lines.append(" | ".join(f"{this_name} \"{op.symbol}\" {next_name} -> op_{op_index}_{index}"
                                    for index, op in enumerate(op)))
            lines.append("\n")
        return "".join(lines)

    def handler(self, *names: str) -> Callable[[Callback], Callback]:
        def wrapper(func: Callback) -> Callback:
            for name in names:
                self.callbacks[name] = func
            return func
        return wrapper

    def handle(self, tree: Tree, stack: Stack) -> Union[Expression, Type]:
        return self.callbacks[tree.data](tree, stack)

    @staticmethod
    def _get_line_column(text:str, position: int) -> Tuple[int, int]:
        line = text.count("\n", 0, position) + 1
        column = position - text[:position].rfind("\n")
        return line, column

    @staticmethod
    def _fix_line_numbers(source: str, tree: Tree):
        """ Calculating line and columns number for each node in the lark tree
            due to an error in lark parser. Line numbers are not correct when
            LALR parser is used. This function call should be removed when the
            future version of lark is available."""
        if isinstance(tree, str):
            return

        if hasattr(tree.meta, "start_pos"):
            tree.meta.line, tree.meta.column = \
                Parser._get_line_column(source, tree.meta.start_pos - 1)
        if hasattr(tree.meta, "end_pos"):
            tree.meta.end_line, tree.meta.end_column = \
                Parser._get_line_column(source, tree.meta.end_pos - 1)

        for child in tree.children:
            Parser._fix_line_numbers(source, child)

    def parse(self, file: Union[IO, str], stack: Stack) -> Iterator[Expression]:
        if not isinstance(file, str):
            file = file.read()
        tree = self.lark.parse(file)
        Parser._fix_line_numbers(file, tree)
        for subtree in tree.children:
            expr = self.handle(subtree, stack)
            if isinstance(expr, Expression):
                yield expr
            else:
                yield from expr


parser = Parser()

for precedence, operators in enumerate(BINARY_OPERATORS):
    for i, operator in enumerate(operators):
        @parser.handler(f"op_{precedence}_{i}")
        def handler(tree: Tree, stack: Stack, op=operator) -> Expression:
            assert len(tree.children) == 2
            expr1: Expression = parser.handle(tree.children[0], stack)
            expr2: Expression = parser.handle(tree.children[1], stack)
            result_type = operator_applicable([expr1.type, expr2.type], op.arguments)
            if result_type is not None:
                return Expression.merge(result_type, op.expression, expr1, expr2)
            else:
                message = f"Operator '{op.symbol}' is not applicable to values"\
                          f" of type {expr1.type} and {expr2.type}"
                raise CompileTimeError(tree, message)


@parser.handler("int_literal")
def handle_int_literal(tree: Tree, _stack: Stack) -> Expression:
    return Expression(tree.line - 1, Integer, str(int(tree.children[0])))


@parser.handler("float_literal")
def handle_float_literal(tree: Tree, _stack: Stack) -> Expression:
    return Expression(tree.line - 1, Real, str(float(tree.children[0])))


@parser.handler("bool_literal_true")
def handle_bool_literal_true(tree: Tree, _stack: Stack) -> Expression:
    return Expression(tree.line - 1, Bool, "True")


@parser.handler("bool_literal_false")
def handle_bool_literal_false(tree: Tree, _stack: Stack) -> Expression:
    return Expression(tree.line - 1, Bool, "False")


@parser.handler("variable")
def handle_variable(tree: Tree, stack: Stack) -> Expression:
    variable = stack.get_variable(tree.children[0])
    if variable is None:
        raise UnresolvedReferenceError(tree.children[0])
    var_name = "var_" + str(variable.number) if variable.number >= 0 else tree.children[0]
    return Expression(tree.line - 1, variable.type, var_name)


@parser.handler("type_int", "type_real", "type_bool", "type_color", "type_side", "type_pattern")
def handle_scalar_type(tree: Tree, _stack: Stack) -> Type:
    return TYPE_NAMES[tree.data]


@parser.handler("type_list", "type_set")
def handle_compound_type(tree: Tree, stack: Stack) -> Type:
    constructor = {"type_list": ListType, "type_set": Set}[tree.data]
    inner_type: Type = parser.handle(tree.children[0], stack)
    return constructor(inner_type)


@parser.handler("var_decl")
def handle_variable_declaration(tree: Tree, stack: Stack) -> List[Expression]:
    var_names: List[str] = []
    index = 0
    while isinstance(tree.children[index], str):
        if str(tree.children[index]) in RESERVED_NAMES:
            raise CompileTimeError(tree, f"`{tree.children[index]}` is not a valid variable name")
        var_names.append(tree.children[index])
        index += 1
    var_type: Type = parser.handle(tree.children[index], stack)
    nums = [stack.add_variable(name, var_type) for name in var_names]

    if index != len(tree.children) - 1:
        value: Expression = parser.handle(tree.children[index + 1], stack)
        assert_type(tree.children[index + 1], value, var_type)
        return [Expression.merge(Void, ["var_" + str(num), " = ", 0], value) for num in nums]
    return [Expression(tree.line - 1, Void, ["var_" + str(num), " = ", var_type.default_value()]) for num in nums]


def flatten(x):
    if isinstance(x, Iterable):
        for y in x:
            yield from flatten(y)
    else:
        yield x


def handle_clause(tree: Tree, stack: Stack) -> List[Expression]:
    if tree.data == "clause":
        stack.add_frame()
        expressions = list(flatten(parser.handle(subtree, stack) for subtree in tree.children))
        stack.pop_frame()
        return expressions
    else:
        stack.add_frame()
        expressions = list(flatten(parser.handle(tree, stack)))
        stack.pop_frame()
        return expressions


@parser.handler("if_expression")
def handle_if_expression(tree: Tree, stack: Stack) -> Expression:
    actions = []
    for j in range(0, len(tree.children) - 1, 2):
        condition = parser.handle(tree.children[j], stack)
        assert_type(tree.children[j], condition, Bool,
                    "Only expression of boolean type can be used as an if condition")
        then_clause = handle_clause(tree.children[j + 1], stack)
        actions.append((condition, then_clause))

    if len(tree.children) % 2 == 1:
        else_clause = handle_clause(tree.children[-1], stack)
    else:
        else_clause = []
    return ConditionExpression(tree.line - 1, actions, else_clause)


@parser.handler("while_expression")
def handle_while_expression(tree: Tree, stack: Stack) -> Expression:
    condition = parser.handle(tree.children[0], stack)
    assert_type(tree.children[0], condition, Bool,
                "Only expression of boolean type can be used as a while condition")
    actions = handle_clause(tree.children[1], stack)
    return WhileLoopExpression(tree.line - 1, condition, actions)


@parser.handler("do_expression")
def handle_do_expression(tree: Tree, stack: Stack) -> Expression:
    condition = parser.handle(tree.children[1], stack)
    assert_type(tree.children[1], condition, Bool,
                "Only expression of boolean type can be used as a do-while condition")
    actions = handle_clause(tree.children[0], stack)
    return DoWhileLoopExpression(tree.line - 1, condition, actions)


@parser.handler("repeat_expression")
def handle_repeat_expression(tree: Tree, stack: Stack) -> Expression:
    iterations = parser.handle(tree.children[0], stack)
    assert_type(tree.children[0], iterations, Integer,
                "Iterations count must be integer")
    actions = handle_clause(tree.children[1], stack)
    return RepeatLoopExpression(tree.line - 1-1 , iterations, actions)


@parser.handler("for_expression")
def handle_repeat_expression(tree: Tree, stack: Stack) -> Expression:
    var_name = tree.children[0]
    range_expression = parser.handle(tree.children[1], stack)
    range_type: CollectionType = range_expression.type

    if not isinstance(range_type, CollectionType):
        raise ValueTypeError(tree.children[1], "For loop range must be a list or a set", None, range_type)

    in_stack = stack.get_variable(var_name)
    if in_stack is not None:
        if in_stack.number >= 0:
            var_name = "var_" + str(in_stack.number)
        assert_type(tree.children[1], range_type.item_type, in_stack.type,
                    f"Value of type {range_type.item_type} cannot be assigned to a for "
                    f"loop iterator of type {in_stack.type}")
        stack.add_frame()
    else:
        stack.add_frame()
        var_num = stack.add_variable(var_name, range_type.item_type)
        var_name = "var_" + str(var_num)

    if tree.children[2].data == "clause":
        actions = [parser.handle(subtree, stack) for subtree in tree.children[2].children]
    else:
        actions = [parser.handle(tree.children[2], stack)]
    stack.pop_frame()
    return ForLoopExpression(tree.line - 1, var_name, range_expression, actions)


@parser.handler("func_call")
def handle_func_call(tree: Tree, stack: Stack):
    def create_arg_list(count):
        if count > 0:
            yield 0
            for index in range(1, count):
                yield ", "
                yield index

    function_name = tree.children[0]
    func_data = stack.get_variable(function_name)
    if func_data is None:
        raise UnresolvedReferenceError(function_name)

    func_type: Function = func_data.type
    if not isinstance(func_type, Function):
        raise ValueTypeError(function_name, f"{function_name} is not a function", None, func_type)

    if func_data.number >= 0:
        function_name = "var_" + str(func_data.number)

    arguments = [parser.handle(x, stack) for x in tree.children[1:]]
    arg_types = [x.type for x in arguments]
    return_type = func_type.takes_arguments(arg_types)
    if return_type is None:
        raise FunctionArgumentsError(tree, function_name, func_type, arg_types)
    if len(arguments) == 0:
        return Expression(tree.line - 1, return_type, [function_name, "()"])
    else:
        return Expression.merge(return_type, [function_name, "(", *create_arg_list(len(arguments)), ")"], *arguments)


@parser.handler("var_assignment")
def handle_var_assignment(tree: Tree, stack: Stack):
    var_name = tree.children[0]
    expression: Expression = parser.handle(tree.children[1], stack)

    if isinstance(var_name, Token):
        var_data = stack.get_variable(var_name)
        if var_data is None:
            raise UnresolvedReferenceError(var_name)
        elif var_data.number < 0:
            raise CompileTimeError(tree, f"Attempting to write to readonly value {var_data}")
        var_type = var_data.type
        result = Expression.merge(Void, ["var_" + str(var_data.number), " = ", 0], expression)
    else:
        # var_name is an array item reference
        var_expression = parser.handle(var_name, stack)
        var_type = var_expression.type
        result = Expression.merge(Void, [0, " = ", 1], var_expression, expression)

    assert_type(tree.children[1], expression, var_type,
                f"Value of type {expression.type} cannot be assigned to a varaible "
                f"of type {var_type}")
    return result


@parser.handler("negation")
def handle_negation(tree: Tree, stack: Stack):
    expr = parser.handle(tree.children[0], stack)
    if not Real.is_assignable(expr.type):
        raise ValueTypeError(tree, "Unary minus can only be applied to integer or real values",
                             Real, expr.type)
    return Expression.merge(expr.type, ["-(", 0, ")"], expr)


@parser.handler("collection_item")
def handle_collection_item(tree: Tree, stack: Stack):
    list_reference: Expression = parser.handle(tree.children[0], stack)
    list_type = list_reference.type
    if not isinstance(list_type, ListType):
        raise ValueTypeError(tree.children[0], f"{list_reference.type} is not a list",
                             None, tree.children)

    index: Expression = parser.handle(tree.children[1], stack)
    assert_type(tree.children[1], index, Integer,
                f"List index must be of type int, not {index.type}")
    return Expression.merge(list_type.item_type, ["(", 0, ")[", 1, "]"], list_reference, index)


@parser.handler("cube_right", "cube_left", "cube_top", "cube_bottom", "cube_front", "cube_back")
def handle_cube_turning(tree: Tree, _stack: Stack):
    side = tree.data[5:]
    return CubeTurningExpression(tree.line - 1, side, 1)


@parser.handler("cube_rotate_right", "cube_rotate_top", "cube_rotate_front")
def handle_cube_rotation(tree: Tree, _stack: Stack):
    side = tree.data[12:]
    return CubeRotationExpression(tree.line - 1, side, False)


@parser.handler("cube_double", "cube_opposite")
def handle_cube_turning_double(tree: Tree, stack: Stack):
    expression = parser.handle(tree.children[0], stack)
    if isinstance(expression, CubeTurningExpression):
        expression.amount = 2 if tree.data == "cube_double" else 3
    else:
        assert isinstance(expression, CubeRotationExpression)
        if tree.data == "cube_double":
            expression.twice = True
        else:
            opposites = ("front", "back"), ("left", "right"), ("top", "bottom")
            for a, b in opposites:
                if expression.side == a:
                    expression.side = b
                    break
                elif expression.side == b:
                    expression.side = a
                    break
    return expression


@parser.handler("cube_turn_range")
def handle_cube_turn_range(tree: Tree, stack: Stack):
    expression: CubeTurningExpression = parser.handle(tree.children[0], stack)
    expression.indices = []

    for i in range(1, len(tree.children)):
        child: Tree = tree.children[i]
        argument = parser.handle(child.children[0], stack)
        assert_type(child.children[0], argument, Integer)

        if child.data == "range_value":
            expression.indices.append(argument)
        elif child.data == "range_open_left":
            expression.indices.append(Expression(tree.line - 1, Void, "..."))
            expression.indices.append(argument)
        elif child.data == "range_open_right":
            expression.indices.append(argument)
            expression.indices.append(Expression(tree.line - 1, Void, "..."))
        else:  # child.data == "range_closed"
            expression.indices.append(argument)
            expression.indices.append(Expression(tree.line - 1, Void, "..."))
            argument_2 = parser.handle(child.children[1], stack)
            assert_type(child.children[1], argument_2, Integer)
            expression.indices.append(argument_2)
    return expression


@parser.handler("cube_color_reference")
def handle_dereference(tree: Tree, stack: Stack):
    side_expression: Expression = parser.handle(tree.children[0], stack)
    assert_type(tree.children[0], side_expression, Side)

    index1: Expression = parser.handle(tree.children[1], stack)
    assert_type(tree.children[1], index1, Integer, "Cube side indices must be integers")

    index2: Expression = parser.handle(tree.children[2], stack)
    assert_type(tree.children[2], index2, Integer, "Cube side indices must be integers")

    return Expression.merge(Color, ["cube_get_color(", 0, ", ", 1, ", ", 2, ")"],
                            side_expression, index1, index2)


@parser.handler("cube_instruction")
def handle_cube_instruction(tree: Tree, stack: Stack):
    return [parser.handle(m, stack) for m in tree.children]


@parser.handler("pattern")
def handle_pattern(tree: Tree, _stack: Stack):
    colors = {"G": "green", "R": "red", "O": "orange", "Y": "yellow",
              "W": "white", "B": "blue", "-": "None"}

    text = tree.children[0][1:-1].split("/")
    pattern_lines = []
    for line in text:
        pattern_line = []
        if len(line) != len(text[0]):
            raise CompileTimeError(tree, "Inconsistent line length in pattern literal")
        for char in line:
            if char in colors:
                pattern_line.append(colors[char])
            else:
                pattern_line.append('"' + char + '"')
        pattern_lines.append("[" + ", ".join(pattern_line) + "]")

    pattern_array = ', '.join(pattern_lines)
    return Expression(tree.line - 1, Pattern, [f"Pattern([{pattern_array}])"])


@parser.handler("orient_params")
def handle_orient_params(tree: Tree, stack: Stack):
    assert len(tree.children) % 2 == 0
    expression = ["orient("]
    merging = []
    previous_keys = set()
    patterns_present = False

    for j in range(0, len(tree.children), 2):
        key = tree.children[j]
        expression.append(key + "=")
        expression.append(j // 2)
        argument = parser.handle(tree.children[j + 1], stack)
        merging.append(argument)
        if key in previous_keys:
            raise CompileTimeError(key, f"Key {key} has already been specified")
        elif key == "keeping":
            assert_type(tree.children[j + 1], argument, Side)
        else:
            patterns_present = True
            assert_type(tree.children[j + 1], argument, Pattern)
        previous_keys.add(key)
        expression.append(", ")

    expression[-1] = ")"
    if not patterns_present:
        raise CompileTimeError(tree, "No side patterns are present")
    return Expression.merge(Bool, expression, *merging)


@parser.handler("func_decl")
def handle_function_declaration(tree: Tree, stack: Stack):
    func_name = tree.children[0]
    argument_names: List[str] = []
    argument_types: List[Type] = []
    index = 1
    while tree.children[index].data == "argument":
        arg_node = tree.children[index]
        argument_names.append(arg_node.children[0])
        type = parser.handle(arg_node.children[1], stack)
        argument_types.append(type)
        index += 1

    if tree.children[index].data == "clause":
        return_type = Void
    else:
        return_type = parser.handle(tree.children[index], stack)
    inner_stack = stack.create_inner(return_type)
    for name, type in zip(argument_names, argument_types):
        inner_stack.add_variable(name, type)

    func_type = Function((argument_types, return_type))
    var_num = stack.add_variable(func_name, func_type)
    clause = handle_clause(tree.children[-1], inner_stack)
    return FunctionDeclarationExpression(tree.line - 1, f"var_{var_num}", func_name, return_type,
                                         [f"var_{i}" for i in range(len(argument_names))], clause)


@parser.handler("return_statement")
def handle_return_statement(tree: Tree, stack: Stack):
    if stack.context_return_type is None:
        raise CompileTimeError(tree, "`return` cannot be used outside of the function body")

    if len(tree.children) > 0:
        value = parser.handle(tree.children[0], stack)
        if not stack.context_return_type.is_assignable(value.type):
            raise ValueTypeError(tree, f"Value of type {stack.context_return_type} is expected to be returned",
                                 value.type, stack.context_return_type)
        return Expression.merge(Void, ["return ", 0], value)
    else:
        if stack.context_return_type == Void:
            return Expression(tree.line - 1, Void, ["return"])
        else:
            raise CompileTimeError(tree, "A value must be returned from this function")


@parser.handler("noop_expression")
def handle_noop_expression(tree: Tree, _stack: Stack):
    return Expression(tree.line - 1, Void, "pass")



from typing import Callable, Optional, List, Union
from collections import deque
import sys

from .actions import Turn, Action, Rotate
from .compiler import types
from .cube import Cube
from .orientation import Orientation, Side, Color
from .pattern import Pattern
from .stdlib import Library
from .execution.rt_error import TerminateExecutionError


class CubeRuntime:

    SIDE_NAMES = {
        "front": Side.FRONT, "back": Side.BACK,
        "left": Side.LEFT, "right": Side.RIGHT,
        "top": Side.TOP, "bottom": Side.BOTTOM
    }

    COLOR_NAMES = {
        "red": Color.RED, "green": Color.GREEN, "blue": Color.BLUE,
        "white": Color.WHITE, "yellow": Color.YELLOW, "orange": Color.ORANGE
    }

    EXPORTED_FUNCTIONS = [
        ("push_orientation", "push_orientation", [], types.Void),
        ("pop_orientation", "pop_orientation", [], types.Void),
        ("suspend_rotations", "suspend_rotations", [], types.Void),
        ("resume_rotations", "resume_rotations", [], types.Void),
        ("exit", "perform_exit", [], types.Void),
        ("print", "debug_print", [types.T, ...], types.Void)
    ]

    def __init__(self, cube: Cube, orientation: Orientation,
                 callback: Callable[[Action], None],
                 done_callback: Callable[[], None]):
        self.cube = cube
        self.orientation = orientation
        self.orientations_stack = deque()
        self.suspended_orientation: Optional[Orientation] = None
        self.callback = callback
        self.done_callback = done_callback

        self.functions = Library()
        for name, local_name, argument_types, return_type in CubeRuntime.EXPORTED_FUNCTIONS:
            self.functions.add_function(name, getattr(self, local_name),
                                        argument_types, return_type)

        self.functions.exec_globals["cube_turn"] = self.perform_turn
        self.functions.exec_globals["cube_rotate"] = self.perform_rotate
        self.functions.exec_globals["cube_get_color"] = self.get_color
        self.functions.exec_globals["orient"] = self.perform_orient
        self.functions.exec_globals["Pattern"] = Pattern

        for name, side in CubeRuntime.SIDE_NAMES.items():
            self.functions.add_value(name, types.Side, side)
        for name, color in CubeRuntime.COLOR_NAMES.items():
            self.functions.add_value(name, types.Color, color)

    def yield_action(self, action: Action) -> None:
        if self.suspended_orientation is None:
            self.callback(action)
        elif isinstance(action, Turn):
            self.callback(action.from_orientation(self.orientation,
                                                  self.suspended_orientation))

    def debug_print(self, *args):
        print(*args, file=sys.stderr)

    def perform_orient(self, *args, **kwargs) -> bool:
        new_orientation = self.cube.orient(self.orientation, *args, **kwargs)
        if new_orientation is not None:
            actions = Rotate.from_turn_steps(self.orientation.turns_to(new_orientation))
            for action in actions:
                self.yield_action(action)
            self.orientation = new_orientation
            return True
        return False

    def suspend_rotations(self):
        self.suspended_orientation = self.orientation

    def resume_rotations(self):
        turns = self.suspended_orientation.turns_to(self.orientation)
        self.suspended_orientation = None
        for action in Rotate.from_turn_steps(turns):
            self.yield_action(action)

    def push_orientation(self):
        self.orientations_stack.append(self.orientation)

    def pop_orientation(self):
        new_orientation = self.orientations_stack.pop()
        turns = self.orientation.turns_to(new_orientation)
        for action in Rotate.from_turn_steps(turns):
            self.yield_action(action)
        self.orientation = new_orientation

    def perform_turn(self, side: Side, amount: int,
                     indices: List[Union[int, type(Ellipsis)]]):
        action = Turn(side, indices, amount)
        self.orientation = action.perform(self.cube, self.orientation)
        self.yield_action(action)

    def perform_rotate(self, side: Side, twice: bool):
        action = Rotate(side, twice)
        self.orientation = action.perform(self.cube, self.orientation)
        self.yield_action(action)

    def perform_exit(self):
        raise TerminateExecutionError()

    def get_color(self, side: Side, i: int, j: int):
        if side == Side.FRONT:
            local_orientation = self.orientation
        elif side == Side.LEFT:
            local_orientation = self.orientation.to_left
        elif side == Side.RIGHT:
            local_orientation = self.orientation.to_right
        elif side == Side.BACK:
            local_orientation = self.orientation.to_left.to_left
        elif side == Side.TOP:
            local_orientation = self.orientation.to_top
        else:
            local_orientation = self.orientation.to_bottom
        return self.cube.get_side(local_orientation).colors[i, j]

    def finished(self):
        self.done_callback()



from setuptools import setup, find_packages
from pathlib import Path
import re


with open(str(Path(__file__).parents[0] / "cubelang" / "__init__.py")) as f:
    init_file = f.read()
    metadata = dict(re.findall(r"^__([a-z_]+)__\s*=\s*\"(.*)\"$", init_file,
                               re.MULTILINE))


setup(
    name='CubeLang',
    version=metadata["version"],
    packages=["cubelang", "cubelang.cli", "cubelang.compiler", "cubelang.execution", 
              "cubelang.postprocessing", "cubelang.stdlib"],
    url='https://github.com/poletaevvlad/CubeLang',
    license='MIT',
    author=metadata["version"],
    author_email=metadata["author_email"],
    description="Domain-specific programming language made for solving twisting cube puzzles",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        "lark-parser",
        "termcolor"
    ],
    entry_points={
        "console_scripts": [
            "cubelang = cubelang.cli.entry:main",
            "cubelang-scramble = cubelang.scrambler:main"
        ]
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Other Audience",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Games/Entertainment :: Puzzle Games",
    ]
)



from termcolor import colored
from typing import IO, Iterable, Tuple, Optional, List
import textwrap
from ..compiler.types import Function, Void, Type
from ..compiler.code_map import CodeMap
from ..execution import RuntimeError
from ..execution.executor import ITracebackWriter


class ErrorsOutput(ITracebackWriter):
    def __init__(self, stream: IO, use_color: bool = False):
        self.stream = stream
        self.max_width = 80
        self.use_color = use_color
        self.line_number_margin = 2
        self.text_indent = 4

    def _echo(self, text: str, color: str = None, nl: bool = False) -> None:
        if self.use_color and color is not None:
            self.stream.write(colored(text, color))
        else:
            self.stream.write(text)
        if nl:
            self.stream.write("\n")

    def _write_lines(self, line_num: str, text: str) -> \
            Iterable[Tuple[int, int]]:
        offset = 0
        width = self.max_width - self.line_number_margin - len(line_num)
        while offset < len(text):
            if offset == 0:
                self._echo(line_num + " " * self.line_number_margin, "cyan")
            else:
                self._echo(" " * (len(line_num) + self.line_number_margin))
            substring = text[offset: offset + width]
            self._echo(substring, nl=True)
            yield offset, offset + len(substring)
            offset += len(substring)

    def _write_lines_with_error(self, line_num: str, text: str,
                                start_col: int, end_col: int) -> None:
        for first, last in self._write_lines(line_num, text):
            if start_col >= last or end_col < first:
                continue
            error_start = max(first, start_col) - first
            error_end = min(last, end_col + 1) - first
            self._echo(" " * (error_start + self.line_number_margin + len(line_num)))
            self._echo("^" * (error_end - error_start), "red", nl=True)

    def _write_lines_all(self, line_num: str, text: str) -> None:
        for _, _ in self._write_lines(line_num, text):
            continue

    def display_code(self, code: str, start_line: int, start_column: int,
                     end_line: int, end_column: int) -> None:
        line_num_width = len(str(end_line))
        code = code.split("\n")

        def write(ln: int) -> None:
            self._write_lines_all(f"{ln + 1:{line_num_width}}", code[ln])

        if end_line - start_line >= 5:
            for line in range(start_line, start_line + 2):
                write(line)
            self._echo("...", nl=True)
            for line in range(end_line - 1, end_line + 1):
                write(line)
        elif end_line != start_line:
            for line in range(start_line, end_line + 1):
                write(line)
        else:
            line_num = f"{start_line + 1:{line_num_width}}"
            self._write_lines_with_error(line_num, code[start_line],
                                         start_column, end_column)

    def write_function_overloads(self, func_name: str, function: Function):
        self._echo("Function arguments:", nl=True, color="yellow")
        for overload in function.overloads:
            indent = self.text_indent + len(func_name) + 1
            offset = indent
            self._echo(" " * self.text_indent + func_name + "(")
            for i, argument in enumerate(overload.arguments):
                string = str(argument)
                if i < len(overload.arguments) - 1:
                    string += ", "
                if offset + len(string) < self.max_width + 1:
                    self._echo(string)
                    offset += len(string)
                else:
                    self._echo("", nl=True)
                    self._echo(" " * indent + string)
                    offset = indent + len(string)
            if overload.return_type == Void:
                ending = ")"
            else:
                ending = f"): {overload.return_type}"
            if offset + len(ending) > self.max_width:
                self._echo("", nl=True)
                self._echo(" " * (indent - 1) + ending, nl=True)
            else:
                self._echo(ending, nl=True)
        self._echo("", nl=True)

    def write_supplied_arguments(self, arguments: List[Type]) -> None:
        self._echo("Supplied arguments:", nl=True, color="yellow")
        self._echo(" " * self.text_indent)
        offset = self.text_indent
        for i, argument in enumerate(arguments):
            string = str(argument)
            if i < len(arguments) - 1:
                string += ", "
            if offset + len(string) > self.max_width:
                self._echo("", nl=True)
                self._echo(" " * self.text_indent + string)
                offset = self.text_indent + len(string)
            else:
                self._echo(string)
                offset += len(string)
        self._echo("\n\n")

    def write_error(self, message: str,
                    line: Optional[int] = None,
                    column: Optional[int] = None):
        header = ["[error"]
        if line is not None or column is not None:
            header.append(" at")
            if line is not None:
                header.append(" line ")
                header.append(str(line + 1))
            if column is not None:
                header.append(" column ")
                header.append(str(column + 1))
        header.append("]")

        self._echo("".join(header), "red", nl=True)

        prefix = " " * self.text_indent
        for line in textwrap.wrap(message, self.max_width - self.text_indent):
            self._echo(prefix + line, nl=True)
        self._echo("", nl=True)

    def print_traceback(self, e: RuntimeError, code_map: CodeMap):
        self._echo("[runtime error]\n", "red", nl=True)
        for line in textwrap.wrap(str(e), self.max_width - self.text_indent):
            self._echo(" " * self.text_indent + line, nl=True)
        self._echo("\n")
        self._echo("Stack trace:", nl=True, color="yellow")
        for func, line in e.stack_entries:
            line_number = "?" if line is None else code_map[line] + 1
            self._echo(" " * self.text_indent + f"line {line_number:<5}")
            if func is None:
                self._echo("\n")
            else:
                self._echo(f": {func}", nl=True)



from .base import PostprocessorBase
from ..actions import Action


class FormattingPostprocessor(PostprocessorBase[Action, str]):
    def process(self, value: Action):
        self._return(str(value))
